#!/usr/bin/env python

# Version 0.1
# Matt Olm
# mattolm@berkeley.edu
# 02.15.19
# https://biotite.berkeley.edu/j/user/mattolm/notebooks/OtherProjects/OneOffs/AlexaPhage_1_writtingProgram.ipynb


import os
import sys
import gzip
import glob
import textwrap
import shutil
import distutils
import argparse
import pandas as pd

from shutil import copyfile
import subprocess
from subprocess import call
from collections import defaultdict

from tqdm import tqdm

from Bio import SeqIO

import drep

def version():
    versionFile = open(os.path.join(drep.__path__[0], 'VERSION'))
    return versionFile.read().strip()

__author__ = "Matt Olm"
__license__ = "MIT"
__version__ = version()


class abstractCommand():
    '''
    abstract class that makes other commands implement these methods
    '''
    def excecute(self):
        raise NotImplementedError

    def verify(self):
        raise NotImplementedError

class ANImCommand(abstractCommand):
    '''
    Hold all of the info necessary to run ANIm
    '''

    def __init__(self, querry=None, reference=None, out_dir=None, settings='BASIC'):
        '''
        Declare all settings
        '''
        # Load settings
        self.config = get_default_args(settings)

        # Store parameters
        self.config['querry'] = querry
        self.config['reference'] = reference
        self.config['out_dir'] = out_dir

        # Find program executable
        self.config['exe'] = find_executable('nucmer')

        # Set that the command has not been run
        self.run = False

    def verify(self):
        '''
        Autocomplete when possbile; return True if this command will likely run without problem
        '''

        # If any arguments are missing, throw exception
        for arg in self.config:
            if self.config[arg] is None:
                raise AttributeError("{0} must be set".format(arg))

        # Make sure required files exist
        for g in ['querry', 'reference']:
            if not os.path.isfile(self.config[g]):
                raise AttributeError("{0} doesn't exist ({1})".format(g,
                    self.config[g]))

        # Make sure output directory exists; try and make it if not
        for d in ['out_dir']:
            if not os.path.isdir(self.config[d]):
                os.makedirs(self.config[d])

        # Get the name of the file
        self.config['prefix'] = self.gen_prefix()

        return True

    def excecute(self, dry=False):
        ''' Run command and return a touple of (ani, coverage)
        '''
        assert self.verify()

        # 1) call the program
        cmd = gen_mummer_cmd(**self.config)
        #print(' '.join(cmd))
        if not dry:
            code = call(cmd)
            if code != 0:
                print("!!! nucmer failed with exit code {0}. Im going to crash now !!!".format(code))
                print(cmd)
                raise Exception("!!! nucmer failed with exit code {0}. Im going to crash now !!!".format(code))

        # 2) Mark that this command has been run
        self.run = True

        # 3) make sure command was successful
        assert os.path.isfile(get_delta_loc(self.gen_prefix()))
        return True

    def parse_results(self, method='simple', **kwargs):
        ''' parse results of the mummer run
        simple = (ani, coverage)
        '''
        # Make sure the command has been run
        if not self.run:
            raise AttributeError('Command has not been run')

        if method == 'simple':
            # Get the organism lengths
            qlen = fasta_length(self.config['querry'])
            rlen = fasta_length(self.config['reference'])

            # Parse the reults
            delta = get_delta_loc(self.config['prefix'])
            return get_ani_simple(delta, qlen, rlen, **kwargs)

        elif method == 'snps':
            return snps_from_delta(get_delta_loc(self.config['prefix']))

        elif method == 'snp_splits':
            # Get the filtered delta location
            f_delta = filter_delta(get_delta_loc(self.config['prefix']))

            # Get the snps location
            snp_loc = get_snps(f_delta)

            # Call the splits
            snp_splits = calc_snp_splits(f_delta, snp_loc)
            return snp_splits

    def gen_prefix(self):
        f = "{0}_vs_{1}".format(os.path.basename(self.config['reference']),
            os.path.basename(self.config['querry']))
        return os.path.join(self.config['out_dir'], f)

    def __str__(self):
        ''' Show the command parameters '''
        return pprint.pformat(self.config)

def gen_mummer_cmd(**kwargs):
    '''
    from a dictionary of arguments, return the ANIm command as an array of strings
    '''

    cmd = [kwargs['exe'],'--' + kwargs['method'],'-p',kwargs['prefix'], '-c', \
        kwargs['c'], '-g', kwargs['maxgap'], '-t', str(kwargs['p'])]

    if kwargs['noextend'] == 'True':
        cmd.append('--noextend')
    cmd += [kwargs['reference'],kwargs['querry']]

    return cmd

def gen_snp_cmd(**kwargs):
    '''
    from a dictionary of arguments, return the show-snp command
    '''
    snp_exe = kwargs.get('snp_exe')
    delta = kwargs.get('delta')
    snp_loc = kwargs.get('snp_loc', get_snp_loc(delta))

    cmd = [snp_exe, '-T', delta, '>', snp_loc]
    return ' '.join(cmd)

def gen_filter_cmd(**kwargs):
    '''
    from a dictionary of arguments, return the delta-filter command
    '''
    filter_exe = kwargs.get('filter_exe')
    delta = kwargs.get('delta')
    filtered_loc = kwargs.get('filtered_loc', get_filtered_loc(delta))

    cmd = [kwargs['filter_exe'], '-r', '-q', delta, '>', filtered_loc]

    return ' '.join(cmd)

def get_ani_simple(delta, qlen, rlen, **kwargs):
    ''' return (ani, coverage) from the delta file
    '''

    aln_length, sim_errors = parse_delta(delta)
    ani = 1 - (sim_errors / aln_length)
    cov = max((aln_length/qlen), aln_length/rlen)

    return (ani, cov)

def parse_delta(filename):
    '''
    parse the delta file and return the raw results
    Returns (alignment length, similarity errors) tuple from passed .delta.
    - filename - path to the input .delta file
    Extracts the aligned length and number of similarity errors for each
    aligned uniquely-matched region, and returns the cumulative total for
    each as a tuple.
    '''

    aln_length, sim_errors = 0, 0
    for line in [l.strip().split() for l in open(filename, 'rU').readlines()]:
        if line[0] == 'NUCMER' or line[0].startswith('>'):  # Skip headers
            continue
        # We only process lines with seven columns:
        if len(line) == 7:
            aln_length += abs(int(line[1]) - int(line[0]))
            sim_errors += int(line[4])
    return aln_length, sim_errors

def parse_delta_extra(filename):
    '''
    Return a pandas table of the details of a .delta file
    '''

    Table = {'r_start':[],'r_end':[], 'r_aln_length':[], \
             'q_start':[],'q_end':[], 'q_aln_length':[], \
             'total_errors':[], 'sim_errors':[],\
             'r_scaff':[],'q_scaff':[],'filename':[],\
             'reference':[],'querry':[]}

    for line in [l.strip().split() for l in open(filename, 'r').readlines()]:
        # Skip top part
        if len(line) == 2:
            reference = line[0]
            querry = line[1]

        # Read header
        elif line[0].startswith('>'):
            r_scaff = line[0][1:]
            q_scaff = line[1]

            r_len = line[2]
            q_len = line[3]

        # Get alignment of alignment lines
        elif len(line) == 7:
            r_start = int(line[0])
            r_end = int(line[1])

            q_start = int(line[2])
            q_end = int(line[3])

            total_errors = int(line[4])
            sim_errors = int(line[5])

            # should always be 0
            assert line[6] == '0'

            Table['r_start'].append(r_start)
            Table['q_start'].append(q_start)
            Table['r_end'].append(r_end)
            Table['q_end'].append(q_end)
            Table['r_aln_length'].append(abs(r_end - r_start))
            Table['q_aln_length'].append(abs(q_end - q_start))
            Table['total_errors'].append(total_errors)
            Table['sim_errors'].append(sim_errors)
            Table['r_scaff'].append(r_scaff)
            Table['q_scaff'].append(q_scaff)
            Table['filename'].append(filename)
            Table['reference'].append(reference)
            Table['querry'].append(querry)

    ddb = pd.DataFrame(Table)

    # Optimize dataframe size
    for cat in ['filename', 'reference', 'querry', 'r_scaff', 'q_scaff']:
        ddb[cat] = ddb[cat].astype('category')

    for num in ['r_start', 'q_start', 'r_end', 'q_end', 'r_aln_length',
            'q_aln_length', 'total_errors', 'sim_errors']:
        ddb[num] = ddb[num].astype(int)

    return ddb

def snps_from_delta(delta, snp_exe=False, filter_exe=False):
    '''
    make a .snp file from a delta file
    first filter, then get snps
    '''
    # filter the delta file
    filter_loc = filter_delta(delta)

    # call snps on filtered_delta
    snp_loc = get_snps(filter_loc)

    return parse_snp(snp_loc)

def filter_delta(delta, filter_exe=False):
    '''
    filter the .delta file
    '''
    if filter_exe is False:
        filter_exe = find_executable('delta-filter')

    filter_loc = get_filtered_loc(delta)
    cmd = gen_filter_cmd(delta=delta, filter_exe=filter_exe)
    #print(cmd)
    call(cmd, shell=True)

    return filter_loc

def get_snps(delta, snp_exe=False):
    '''
    call snps from delta file
    '''

    if snp_exe is False:
        snp_exe = find_executable('show-snps')

    snp_loc = get_snp_loc(delta)
    cmd = gen_snp_cmd(delta=delta, snp_exe=snp_exe)
    #print(cmd)
    call(cmd, shell=True)

    return snp_loc

def parse_snp(loc):
    '''
    return file as pandas dataframe
    '''
    h = ['rl', 'rb', 'qb', 'ql', 'buff', 'dist', 'R', 'Q', 'FRM', 'TAGS', 'rscaff', 'qscaff']
    x = pd.read_table(loc, skiprows=4, header=None)
    x.columns = h

    x['ql'] = x['ql'].astype(int)
    x['rl'] = x['rl'].astype(int)
    x['buff'] = x['buff'].astype(int)
    x['dist'] = x['dist'].astype(int)
    x['R'] = x['R'].astype(int)
    x['Q'] = x['Q'].astype(int)
    x['FRM'] = x['FRM'].astype(int)
    x['TAGS'] = x['TAGS'].astype(int)
    x['rscaff'] = x['rscaff'].astype('category')
    x['qscaff'] = x['qscaff'].astype('category')

    return x

def filter_snp(db):
    '''
    filter the input snp db such that indels only count as one line
    '''
    d = db.drop_duplicates(subset=['ql', 'qscaff'])
    d = d.drop_duplicates(subset=['rl', 'rscaff'])
    return d

def calc_snp_splits(delta_file, snp_file, **kwargs):
    # 0) get arguments
    split_len = kwargs.get('split_len', 1000)

    # 1) parse both files
    delta_db = parse_delta_extra(delta_file)
    snp_db = parse_snp(snp_file)

    # 2) calculate split identities
    split_dic = make_split_dic(delta_db, snp_db, split_len)

    return split_dic

def make_split_dic(delta_db, snp_db, split_len):
    '''
    Return a table of splits
    It is essential that each part of the reference only be aligned to one region
    of the querry. This is because I have a list of SNPs, but I don't know what
    alignment they're to. So if a SNP is in the reference sequence in only one
    of two alignments, I would have no way of knowing that, and will mark that
    reference location as a SNP in each alignment it goes to.
    '''
    Sdb = pd.DataFrame()

    snp_db = filter_snp(snp_db)
    for scaff in delta_db['r_scaff'].unique():
        table = {'scaffold':[], 'start':[], 'end':[], 'snps':[]}
        found_locs = [] # keep track of snps from this scaffold
        ddb = delta_db[delta_db['r_scaff'] == scaff].sort_values('r_start')
        locs = sorted(snp_db['rl'][(snp_db['rscaff'] == scaff)].tolist())

        # for every alignment in this scaffold...
        for start, end in zip(ddb['r_start'], ddb['r_end']):
            assert start < end
            cur_locs = [] # keep track of all snps in this alignment

            # set up the modulo dictionary
            split2snps = {}

            # for every snp in this scaffold...
            for x in locs:

                # if the snps is in the current alignment...
                if start <= x < end:
                    cur_locs.append(x)
                    found_locs.append(x)

                    # add it to the split
                    sp = get_split(start, x, split_len)
                    split2snps.setdefault(sp, []).append(x)

            # add the results from this alignment to the table
            for split, snps in split2snps.items():
                st, ed = get_range_from_split(start, split, split_len, end)
                table['scaffold'].append(scaff)
                table['start'].append(st)
                table['end'].append(ed)
                table['snps'].append(len(snps))

        # make sure every snp in the scaffold was added to an alignment and split
        sdb = pd.DataFrame(table)
        assert len(found_locs) == len(locs) == sdb['snps'].sum()

        # add to the table
        Sdb = pd.concat([Sdb, sdb], ignore_index=True)

    Sdb = _finalize_table(Sdb)
    return Sdb

def find_executable(prog):
    try:
        exe = distutils.spawn.find_executable(prog)
    except:
        exe = False
    return exe

def get_split(start, x, split_len):
    '''
    return the split to which x belongs
    '''
    return int((x-start)/split_len)

def get_range_from_split(start, sp, split_len, end):
    '''
    return the range of values for this split
    '''
    st = start + (sp * split_len)
    ed = min(st + split_len, end)
    return st, ed

def _finalize_table(sdb):
    sdb['length'] = sdb['end'] - sdb['start']
    sdb['ani'] = [(l-s)/l for s, l in zip(sdb['snps'], sdb['length'])]

    sdb['scaffold'] = sdb['scaffold'].astype('category')
    sdb['end'] = sdb['end'].astype(int)
    sdb['start'] = sdb['start'].astype(int)
    sdb['snps'] = sdb['snps'].astype(int)
    sdb['length'] = sdb['length'].astype(int)

    return sdb

def get_delta_loc(prefix):
    '''
    return the .delta filename based on the prefix in the original command
    '''
    return prefix + '.delta'

def get_snp_loc(delta):
    '''
    return the .snp filename based on the prefix in the original command
    '''
    return delta + '.snp'

def get_filtered_loc(delta):
    '''
    return the .snp filename based on the prefix in the original command
    '''
    return delta + '.filtered'

def main(args):
    '''
    Main entry point
    '''
    if not os.path.isdir(args.outdir):
        os.mkdir(args.outdir)
        # print("outdir has to exist. You gave me {0}, which does not".format(
        # args.outdir))
        # sys.exit()

    dereplicate_scaffolds(args.fasta_files, args.outdir, **vars(args))

def get_default_args(settings):
    '''
    Load default arguments from file and return dictionary of results
    '''
    if settings == 'BASIC':
        return {'c': '65', 'noextend': 'False', 'maxgap': '90', 'method': 'mum'}
    else:
        return {}


def dereplicate_scaffolds(fastas, output_folder, **kwargs):
    '''
    De-replicate several lists of scaffolds

    Arguments:
        fastas = List of .fasta files to be de-replicated on a scaffold level, in priority order
        output_folder = Folder to store output files
    '''
    verbose = kwargs.get('verbose', False)

    # Do some quick validating
    if output_folder[-1] != '/':
        output_folder = output_folder + '/'
    assert os.path.isdir(output_folder)

    if verbose:
        print("Step 1- Load scaffold to length for .fasta files")
    scaffDb = load_fastas(fastas)

    assert len(fastas) >= 1
    selfCompairson = False
    if len(fastas) == 1:
        kwargs['IgnoreSameScaffolds'] = True
        selfCompairson = True

        # Duplicate the scaffold
        scaffDb2 = scaffDb.copy()
        scaffDb2['priority'] = 1
        scaffDb = pd.concat([scaffDb, scaffDb2]).reset_index(drop=True)

    if verbose:
        print("Step 2- Run comparisons and dereplicate scaffolds")
    final_file, RMdb, Sdb = compare_scaffolds(scaffDb, output_folder, **kwargs)

    if verbose:
        print("Step 3- Generate report")
    gen_report(output_folder, final_file, RMdb, Sdb, scaffDb, verbose=verbose,
                selfCompairson=selfCompairson)

def load_fastas(fastas):
    '''
    Create a pandas dataframe of scaffold to length
    '''
    Sdb = pd.DataFrame()
    for i, fasta in enumerate(fastas):
        db = load_scaffold2length(fasta)
        db['loc'] = fasta
        db['priority'] = i
        db['sort_name'] = os.path.basename(fasta)
        Sdb = Sdb.append(db)

    Sdb['loc'] = Sdb['loc'].astype('category')
    Sdb['sort_name'] = Sdb['sort_name'].astype('category')
    Sdb['priority'] = Sdb['priority'].astype(int)
    Sdb['length'] = Sdb['length'].astype(int)

    return Sdb

def compare_scaffolds(scaffDb, output_folder, **kwargs):
    '''
    De-replicate scaffolds

    This follows a couple of steps:
    '''
    verbose = kwargs.get('verbose')

    if verbose:
        print("Step 2.1- Prepare a directory")
    temp_folder = os.path.join(output_folder + 'comparisons/')
    if not os.path.exists(temp_folder):
        os.mkdir(temp_folder)

    if verbose:
        print("Step 2.2- Run comparisons")

    # Current is the .fasta that you differ to. Initially this is the first .fasta in the list
    indecies = sorted(list(scaffDb['priority'].unique()))[1:]
    current = scaffDb[scaffDb['priority'] == 0]['loc'].unique()[0]

    # Keep track of removed scaffolds
    RMdb = pd.DataFrame()

    for i in indecies:
        # Scaffolds that are in new and current will be removed from new
        new = scaffDb[scaffDb['priority'] == i]['loc'].unique()[0]

        if verbose:
            print('comparing {0} and {1}'.format(current, new))
        Ddb, Sdb = compare_wrapper(current, new, scaffDb, temp_folder, **kwargs)

        if verbose:
            print('De-replicating scaffolds')

        # Make a new file where scaffolds that were in new and current are removed from new
        newfile = os.path.join(temp_folder, "deReplication_v{0}.fasta".format(i))
        rmdb = reconsile(Sdb, current, new, newfile, temp_folder, **kwargs)

        # Keep track of what was removed
        RMdb = RMdb.append(rmdb)

        current = newfile

    return current, RMdb, Sdb

def gen_report(output_folder, final_file, RMdb, Sdb, scaffDb, verbose=False, selfCompairson=False):
    '''
    Generate report
    '''
    if verbose:
        print('generating report')

    # Save final file
    ffile = os.path.join(output_folder + 'dereplicated_scaffolds.fasta')
    copyfile(final_file, ffile)

    # Save scaffold report
    rfile = os.path.join(output_folder + 'DereplicationInfo.csv')
    RMdb.to_csv(rfile, index=False)

    # Print report
    print("""
    $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
    $$$                                              $$$
    $$$    All finished! Here's your final report    $$$
    $$$                                              $$$
    $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
    """)

    if selfCompairson:
        indecies = [0]
    else:
        indecies = sorted(list(scaffDb['priority'].unique()))

    for i in indecies:
        db = scaffDb[scaffDb['priority'] == i]
        loc = db['loc'].unique()[0]
        ddb = RMdb[RMdb['file_removed_from'] == os.path.basename(loc)]
        print("{0} - {2} of {1} scaffolds removed".format(os.path.basename(loc),
                len(db), len(ddb)))

    print('\nDereplicated scaffolds located at {0}'.format(ffile))
    print('Details on scaffolds removed located at {0}'.format(rfile))

    if selfCompairson:
        # Save the ANI table
        rfile = os.path.join(output_folder + 'Scaffold_ANI_values.csv')
        Sdb.to_csv(rfile, index=False)
        print('Details on scaffolds removed located at {0}'.format(rfile))


def load_scaffold2length(fasta):
    '''
    Return a database of scaffold to length
    '''
    table = defaultdict(list)

    if fasta[-3:] == '.gz':
        with gzip.open(fasta, "rt") as handle:
            for seq_record in SeqIO.parse(handle, "fasta"):
                table['scaffold'].append(seq_record.id)
                table['length'].append(len(seq_record))
    else:
        for seq_record in SeqIO.parse(str(fasta), "fasta"):
            table['scaffold'].append(seq_record.id)
            table['length'].append(len(seq_record))
    return pd.DataFrame(table)

def add_cov_ani(Ddb, s2l):
    '''
    From an unfiltered Delta, report the coverage and ANI of each alignment.

    Based on length of reference sequence
    '''
    # Add coverage and ANI at the alignment level
    Ddb['coverage'] = [x/s2l[y] for x,y in zip(Ddb['r_aln_length'], Ddb['r_scaff'])]
    Ddb['ani'] = [(l -e)/l for l,e in zip(Ddb['r_aln_length'], Ddb['total_errors'])]

    # Add coverage and ANI at the scaffold level
    table = defaultdict(list)
    for rscaff, db in Ddb.groupby('r_scaff'):
        for qscaff in db['q_scaff'].unique():
            d = db[db['q_scaff'] == qscaff]
            if len(d) == 0:
                continue

            #print(qscaff)
            table['r_scaff'].append(rscaff)
            table['q_scaff'].append(qscaff)
            table['r_aln_length'].append(d['r_aln_length'].sum())
            table['total_errors'].append(d['total_errors'].sum())
            table['alignments'].append(len(d))


    Sdb = pd.DataFrame(table)
    Sdb['coverage'] = [x/s2l[y] for x,y in zip(Sdb['r_aln_length'], Sdb['r_scaff'])]
    Sdb['ani'] = [(l -e)/l for l,e in zip(Sdb['r_aln_length'], Sdb['total_errors'])]

    return Ddb, Sdb

def split_into_chunks(file, max_len, temp_folder):
    '''
    Split the file "file" into numerous files of max_len size
    '''
    newfile_base = os.path.join(temp_folder, os.path.basename(file))
    chunks = []

    chunk_number = 0
    records = []
    chunk_len = 0
    for r in SeqIO.parse(file, "fasta"):
        records.append(r)
        chunk_len += len(r)

        if chunk_len > max_len:
            # Save this chunk
            new_loc = newfile_base + '_CHUNK_{0}.fasta'.format(chunk_number)
            SeqIO.write(records, new_loc, "fasta")
            chunks.append(new_loc)

            chunk_len = 0
            chunk_number += 1
            records = []

    # Save the final chunk
    if len(records) > 0:
        new_loc = newfile_base + '_CHUNK_{0}.fasta'.format(chunk_number)
        SeqIO.write(records, new_loc, "fasta")
        chunks.append(new_loc)

    return chunks

def compare_wrapper(current, new, scaffDb, temp_folder, **kwargs):
    '''
    A wrapper around compare that can split and paralellize if needed
    '''
    p = kwargs.get('processes')
    verbose = kwargs.get('verbose')
    max_len = kwargs.get('max_len')

    # Get length of current
    if current in set(scaffDb['loc']):
        clen = scaffDb[scaffDb['loc'] == current]['length'].sum()
    else:
        print("YOU HAVENT CODED THE ABILITY TO SPLIT NEW .FASTAS YET!!!")
        clen = max_len - 1

    # Get length of new
    if new in set(scaffDb['loc']):
        nlen = scaffDb[scaffDb['loc'] == new]['length'].sum()
    else:
        print("YOU HAVENT CODED THE ABILITY TO SPLIT NEW .FASTAS YET!!!")
        nlen = max_len - 1

    # See if either of them need to be split
    if clen > max_len:
        c_chunks = split_into_chunks(current, max_len, temp_folder)
    else:
        c_chunks = [current]

    if nlen > max_len:
        n_chunks = split_into_chunks(new, max_len, temp_folder)
    else:
        n_chunks = [new]

    # Pairwise comparisons
    ddbs = []
    sdbs = []
    pbar = tqdm(desc='Running comparisons: ', total=(len(n_chunks) * len(c_chunks)))
    for c, c_chunk in enumerate(c_chunks):
        for n, n_chunk in enumerate(n_chunks):
            ddb, sdb = compare(c_chunk, n_chunk, scaffDb, temp_folder, p=p, verbose=verbose)
            ddb['c_chunk'] = c
            ddb['n_chunk'] = n
            sdb['c_chunk'] = c
            sdb['n_chunk'] = n
            ddbs.append(ddb)
            sdbs.append(sdb)
            pbar.update(1)

    pbar.close()
    return pd.concat(ddbs).reset_index(drop=True), pd.concat(sdbs).reset_index(drop=True)

def compare(current, new, scaffDb, temp_folder, p=1, verbose=False):
    '''
    Run mummer comparisons between .fasta files
    '''

    #Make mummer command
    delta_name = "{0}-vs-{1}".format(os.path.basename(current), os.path.basename(new))

    destf = "{0}{1}.delta".format(temp_folder, delta_name)
    ANIm_cmd = ANImCommand()
    ANIm_cmd.config['querry'] = current
    ANIm_cmd.config['reference'] = new
    ANIm_cmd.config['out_dir'] = temp_folder
    ANIm_cmd.config['p'] = p
    ANIm_cmd.config['exe'] = shutil.which('nucmer')
    ANIm_cmd.verify()

    #Run original
    cmd = gen_mummer_cmd(**ANIm_cmd.config)
    delta = get_delta_loc(ANIm_cmd.gen_prefix())
    ANIm_cmd.excecute()

    #Run filter
    filter_exe = shutil.which('delta-filter')
    f_result = filter_delta(delta, filter_exe)

    #Load and parse result
    s2l = scaffDb.set_index('scaffold')['length'].to_dict()
    Ddb = parse_delta_extra(f_result)
    if len(Ddb) > 0:
        Ddb['q_scaff'] = Ddb['q_scaff'].astype('category')
        Ddb['r_scaff'] = Ddb['r_scaff'].astype('category')
        Ddb, Sdb = add_cov_ani(Ddb, s2l)

        #Store
        Sdb.to_csv("{0}{1}.summary.csv".format(temp_folder, delta_name), index=False)
        Ddb.to_csv("{0}{1}.allAlignments.csv".format(temp_folder, delta_name), index=False)

    else:
        Sdb = pd.DataFrame({'coverage':[], 'ani':[], 'q_scaff':[], 'r_scaff':[]})

    #Return
    return Ddb, Sdb

def reconsile(Sdb, current, new, newFile, temp_folder, **kwargs):
    '''
    Make a new .fasta file combining current and new, based on Ddb and Sdb
    '''
    IgnoreSameScaffolds = kwargs.get('IgnoreSameScaffolds')
    minCov = kwargs.get('minCov')
    minANI = kwargs.get('minANI')
    verbose = kwargs.get('verbose')

    Rdb = Sdb.copy()
    Rdb = Rdb[(Rdb['coverage'] >= minCov) & (Rdb['ani'] >= minANI)]
    if IgnoreSameScaffolds:
        Rdb = Rdb[Rdb['q_scaff'] != Rdb['r_scaff']]

    # All querry should be keepers
    keepers = set(Rdb['q_scaff'])

    # All ref should be removered
    to_rm = set(Rdb['r_scaff'])

    # Make a new ref
    new_ref = os.path.join(temp_folder, os.path.basename(new) + '.dr')
    write_new(new, new_ref, to_rm, verbose=verbose)

    # Merge with current
    new_current = newFile
    with open(new_current, 'w') as outfile:
        for fname in [current, new_ref]:
            if fname[-3:] == '.gz':
                with gzip.open(fname,'r') as infile:
                    for line in infile:
                        outfile.write(line.decode('utf-8'))
            else:
                with open(fname) as infile:
                    for line in infile:
                        outfile.write(line)

    # Make an ourput to return
    Rdb = Rdb[['r_scaff', 'q_scaff', 'coverage', 'ani']]
    Rdb = Rdb.rename(columns={'q_scaff':'scaffold_retained', 'r_scaff':'scaffold_removed'})
    Rdb['file_removed_from'] = os.path.basename(new)

    return Rdb


def write_new(ori_loc, new_loc, to_rm, verbose=False):
    '''
    Write a new .fasta file removing certain scaffolds
    '''
    records = []
    if ori_loc[-3:] == '.gz':
        with gzip.open(ori_loc, "rt") as handle:
            for r in SeqIO.parse(handle, "fasta"):
                records.append(r)
    else:
        for r in SeqIO.parse(ori_loc, "fasta"):
            records.append(r)

    ori_len = len(records)
    records = (r for r in records if r.id not in to_rm)
    count = SeqIO.write(records, new_loc, "fasta")
    if verbose:
        print("Removed {0} scaffolds from {1}".format(ori_len - count, ori_loc))


class test_dRepScaffolds():
    '''
    Tests
    '''
    def setUp(self):
        self.fastas = glob.glob('/home/mattolm/Bio_scripts/test_suite/test_files/*_min1000.fa') \
        + ['/home/mattolm/Bio_scripts/test_suite/test_files/N5_271_010G1_scaffold_3.fasta',
          '/home/mattolm/Bio_scripts/test_suite/test_files/NZ_BAGX02000038.1.fasta']
        self.test_dir = '/home/mattolm/Bio_scripts/test_suite/test_backend/testdir/'

        if os.path.isdir(self.test_dir):
            shutil.rmtree(self.test_dir)
        os.mkdir(self.test_dir)

    def tearDown(self):
        if os.path.isdir(self.test_dir):
            shutil.rmtree(self.test_dir)

    def run(self):
        self.setUp()
        self.test1()
        self.tearDown()

        self.setUp()
        self.test2()
        self.tearDown()

        self.setUp()
        self.test3()
        self.tearDown()

        self.setUp()
        self.test4()
        self.tearDown()

        print('all clear')

    def test1(self):
        '''
        Test load .fasta files
        '''
        scaffDb = load_fastas(self.fastas)
        assert len(scaffDb['priority'].unique() == 2)

    def test2(self):
        '''
        Test Compare
        '''
        scaffDb = load_fastas(self.fastas)
        Ddb, Sdb = dereplicate_scaffolds(self.fastas, self.test_dir, verbose=True)

        print(Ddb.head())
        print(Sdb.head())

    def test3(self):
        '''
        Test write_new
        '''
        to_rm = ['N5_271_010G1_scaffold_119']
        new_loc = self.test_dir + 'new_file.fasta'

        ori_len = len([r for r in SeqIO.parse(self.fastas[0], "fasta")])
        write_new(self.fastas[0], new_loc, to_rm, verbose=False)
        new_len = len([r for r in SeqIO.parse(new_loc, "fasta")])

        assert new_len + 1 == ori_len

    def test4(self):
        '''
        Test all
        '''
        dereplicate_scaffolds(self.fastas, self.test_dir, verbose=True)


#test_dRepScaffolds().run()

if __name__ == "__main__":
    """ This is executed when run from the command line """
    parser = argparse.ArgumentParser(formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        usage=textwrap.dedent('''\
         \n
         dRep_scaffoldLevel combines many lists of scaffolds into a single,
         dereplicated list. \n\n
         -f Takes the list of .fasta files. The first .fasta file in the list
         will take priority, and all scaffolds in the second file will be removed.
         The third file will then be compared to a combination of the first and second
         files, ect.

         If only one .fasta file is given it will be compared with itself, and
         --IgnoreSameScaffolds will be automatically enabled

         -o Is the directory where all output will be created. Check in the "comparisons" folder
         for raw information

         -minCov Is the minimum alignment coverage between two scaffolds for them
         to be considered the same. minCov and minANI must both be satisfied to remove
         a scaffold

         -minANI Is in the minimum average nucleotide identity (ANI) for two scaffolds
         to be the same
         '''))

    parser.add_argument("-f", "--fasta_files", nargs='*', help='fasta files to dereplicate on a per-scaffold level')
    parser.add_argument("-o", "--outdir", help='output directory')
    parser.add_argument("-c", "--minCov", help='minimum alignment coverage', default=0.5, type=float)
    parser.add_argument("-a", "--minANI", help='minimum alignment coverage', default=0.95, type=float)
    parser.add_argument("-m", "--max_len", help='maximum length (in bp) of each .fasta file. Split .fasta files above this length. Making this number lower will increase run-time and decrease RAM usage', default=1000000000, type=int)
    parser.add_argument("--IgnoreSameScaffolds", help='Dont dereplicate scaffolds with same name', action='store_true', default=False)
    parser.add_argument("-p", "--processes", help='number of processes to use', default=6, type=int)
    parser.add_argument('-d', "--debug", help='print extra output', action='store_true', default=False)

    # Specify output of "--version"
    parser.add_argument(
        "--version",
        action="version",
        version="%(prog)s (version {version})".format(version=__version__))

    args = parser.parse_args()
    main(args)



import glob
import logging
import os
import sys

import pandas as pd

import drep
import drep.d_cluster.cluster_utils
import drep.d_cluster.external
import drep.d_cluster.utils
import drep.d_cluster.greedy_clustering

class genomeChunk():
    """
    This is an object that just holds stuff related to a chunk of genomes for Mash clustering
    """

    def __init__(self, locations, number, sketch_folder, genome_names, no_create=False):
        '''
        Initialize this genome chunk
        '''
        self.name = "chunk_{0}".format(number)
        self.genome_locations = locations
        self.chunk_folder = os.path.join(sketch_folder, self.name)
        self.genome_names = genome_names

        if not no_create:
            if not os.path.exists(self.chunk_folder):
                os.makedirs(self.chunk_folder)

    def gen_sketch_cmds(self, mash_exe, MASH_s):
        cmds = []
        for location, name in zip(self.genome_locations, self.genome_names):
            file = os.path.join(self.chunk_folder, name)
            if not os.path.isfile(file + '.msh'):
                cmd = [mash_exe, 'sketch', location, '-s', str(MASH_s), '-o',
                       file]
                cmds.append(cmd)
        return cmds

    def gen_paste_cmd(self, mash_exe):
        all_file = os.path.join(self.chunk_folder, 'chunk_all.msh')
        cmd = [mash_exe, 'paste', all_file] \
              + glob.glob(os.path.join(self.chunk_folder, '*'))
        self.all_file = all_file
        return cmd

    def gen_dist_cmd(self, mash_exe, mash_folder, p):
        dist_file = os.path.join(mash_folder, '{0}_MASH_table.tsv'.format(self.name))
        self.dist_file = dist_file
        cmd = [mash_exe, 'dist', '-p', str(p), self.all_file, self.all_file, '>', dist_file]
        cmd = ' '.join(cmd)

        return cmd

    def load_mash_table(self):
        Mdb = drep.d_cluster.utils.parse_mash_table(self.dist_file)

        # Filter out those genomes that are in the MASH folder but shouldn't be in Mdb
        gs = set(self.genome_names)
        Mdb = Mdb[Mdb['genome1'].isin(gs)]
        Mdb = Mdb[Mdb['genome2'].isin(gs)]

        # Reorder categories to be correct
        for g in ['genome1', 'genome2']:
            Mdb[g] = Mdb[g].cat.remove_unused_categories()
            Mdb[g] = Mdb[g].cat.reorder_categories(sorted((Mdb[g].unique())), ordered=True)

        self.Mdb = Mdb

    def cluster_mash_table(self, **kwargs):
        if len(self.Mdb) > 1:
            Cdb, cluster_ret = cluster_mash_database(self.Mdb, **kwargs)
            self.Cdb = Cdb
        else:
            self.Cdb = pd.DataFrame({'primary_cluster':[1], 'genome':[self.Mdb['genome1'].tolist()[0]]})

    def get_winning_genomes(self, g2l, **kwargs):
        Cdb = self.Cdb
        Cdb['length'] = Cdb['genome'].map(g2l)
        return self.Cdb.sort_values('length').drop_duplicates(subset=['primary_cluster'], keep='last')['genome'].tolist()

def all_vs_all_MASH(Bdb, data_folder, **kwargs):
    """
    Run MASH pairwise within all samples in Bdb

    Args:
        Bdb: dataframe with genome, location
        data_folder: location to store temporary output files

    Keyword Args:
        MASH_sketch: size of mash sketches
        dry: dont actually run anything
        processors: number of processors to multithread with
        mash_exe: location of mash excutible (will try and find with shutil if not provided)
        groupSize: max number of mash sketches to hold in each folder
        debug: if True, log all of the commands
        wd: if you want to log commands, you also need the wd
    """
    # Set up the mash folder structure
    logdir, MASH_folder, sketch_folder, mash_exe = prepare_mash(data_folder, **kwargs)

    # Set up chunks of genomes
    genome_chunks = prepare_genome_chunks(Bdb, sketch_folder, MASH_folder,  **kwargs)
    if len(genome_chunks) > 1:
        logging.info(f"  Will split genomes into {len(genome_chunks)} groups for primary clustering")

    # Process the chunks individually
    genome_chunks = run_mash_on_genome_chunks(genome_chunks, mash_exe, sketch_folder, MASH_folder, logdir,  **kwargs)

    # If there's only one chunk, we're done here
    if len(genome_chunks) == 1:
        Mdb = genome_chunks[0].Mdb
        Cdb, cluster_ret = cluster_mash_database(Mdb, **kwargs)
        return Mdb, Cdb, cluster_ret

    # If there's multiple chunks, run a second round
    logging.info("  Final step: comparing between all groups")
    return run_second_round_clustering(Bdb, genome_chunks, data_folder, verbose=True, **kwargs)

def prepare_mash(data_folder, **kwargs):
    """
    Make some folders and things
    """
    append = kwargs.get('v2', '')

    # set up logdir
    if ('wd' in kwargs) and (kwargs.get('debug', False) == True):
        logdir = kwargs.get('wd').get_dir('cmd_logs')
    else:
        logdir = False

    # Find mash excutable
    mash_exe = kwargs.get('exe_loc', None)
    if mash_exe == None:
        mash_exe = drep.get_exe('mash')

    # Make a folder to hold this information
    MASH_folder = os.path.join(data_folder, 'MASH_files{0}/'.format(append))
    if not os.path.exists(MASH_folder):
        os.makedirs(MASH_folder)

    # Make a folder in there to store sketches
    sketch_folder = os.path.join(MASH_folder, 'sketches{0}/'.format(append))
    if not os.path.exists(sketch_folder):
        os.makedirs(sketch_folder)

    return logdir, MASH_folder, sketch_folder, mash_exe

def prepare_genome_chunks(Bdb, sketch_folder, MASH_folder, **kwargs):
    groupSize = kwargs.get('primary_chunksize', 5000)
    l2g = Bdb.set_index('location')['genome'].to_dict()

    locations = list(Bdb['location'].unique())
    chunks = [locations[x:x + groupSize] for x in range(0, len(locations), groupSize)]

    genome_chunks = []
    for i, chunk in enumerate(chunks):
        genome_chunks.append(genomeChunk(chunk, i, sketch_folder, [l2g[l] for l in chunk]))

    return genome_chunks

def run_mash_on_genome_chunks(genome_chunks, mash_exe, sketch_folder, MASH_folder, logdir, **kwargs):
    dry = kwargs.get('dry', False)
    p = kwargs.get('processors', 6)
    MASH_s = kwargs.get('MASH_sketch', 1000)
    multi_round = kwargs.get('multiround_primary_clustering', True)

    # Step 1) Create Mash sketches
    cmds = []
    for GC in genome_chunks:
        cmds += GC.gen_sketch_cmds(mash_exe, MASH_s)
    if (not dry) & (len(cmds) > 0):
        drep.thread_cmds(cmds, logdir=logdir, t=int(p))

    # Step 2) Combine MASH sketches within chunks
    cmds = [GC.gen_paste_cmd(mash_exe) for GC in genome_chunks]
    if (not dry) & (len(cmds) > 0):
        drep.thread_cmds(cmds, logdir=logdir, t=int(p))

    # Merge the pasted chunks and make a new genomeChunk if thats what you want
    if (not multi_round) & (len(genome_chunks) > 1):
        cmd, new_gc = drep.d_cluster.utils.merge_genome_chunks(mash_exe, genome_chunks, sketch_folder, MASH_folder)
        genome_chunks = [new_gc]
        drep.run_cmd(cmd, dry, shell=False, logdir=logdir)

    # Step 3) Run Mash on each chunk
    cmds = [GC.gen_dist_cmd(mash_exe, MASH_folder, p) for GC in genome_chunks]
    for j, cmd in enumerate(cmds):
        if not dry:
            if len(cmds) > 1:
                logging.info(f"  Comparing group {j+1} of {len(cmds)}")
            drep.run_cmd(cmd, dry, shell=True, logdir=logdir)

    # Step 4) Load the Mash tables of each chunk
    for GC in genome_chunks:
        GC.load_mash_table()

    return genome_chunks

def run_second_round_clustering(Bdb, genome_chunks, data_folder, **kwargs):
    verbose = kwargs.get('verbose', False)

    kwargs_copy = kwargs.copy()
    kwargs_copy['multiround_primary_clustering'] = False
    kwargs_copy['v2'] = '_v2'

    mdbs = []

    # Step 1) Create a merged Cdb file
    dbs = []
    for gc in genome_chunks:
        gc.cluster_mash_table(**kwargs_copy)
        cdb = gc.Cdb
        cdb['subcluster'] = ["{0}_{1}".format(gc.name, x) for x in cdb['primary_cluster']]
        dbs.append(cdb)

        mdb = gc.Mdb
        mdb['genome_chunk'] = gc.name
        mdbs.append(mdb)

    Cdb = pd.concat(dbs)

    # Step 2) Pick winners
    g2l = Bdb.set_index('genome')['length'].to_dict()
    Cdb['length'] = Cdb['genome'].map(g2l)
    second_round_genomes = Cdb.sort_values('length').drop_duplicates(subset=['subcluster'], keep='last')['genome'].tolist()

    if verbose:
        logging.info(f"Comparing {len(second_round_genomes):,} genomes")

    # Step 3) Run a second round
    logdir, MASH_folder, sketch_folder, mash_exe = prepare_mash(data_folder, **kwargs_copy)
    genome_chunks = prepare_genome_chunks(Bdb[Bdb['genome'].isin(second_round_genomes)], sketch_folder, MASH_folder, **kwargs_copy)
    genome_chunks = run_mash_on_genome_chunks(genome_chunks, mash_exe, sketch_folder, MASH_folder, logdir, **kwargs_copy)

    # Step 4) Get results
    assert len(genome_chunks) == 1

    mdb = genome_chunks[0].Mdb
    mdb['genome_chunk'] = 'v2'
    mdbs.append(mdb)
    Mdb = pd.concat(mdbs).reset_index(drop=True)

    Cdb2, cluster_ret = cluster_mash_database(mdb, **kwargs)
    Cdb2['primary_representitive'] = True

    # Step 5) Merge the new Cdb back in with the old
    del Cdb['primary_cluster']
    Cdb = pd.merge(Cdb, Cdb2, on='genome', how='outer')
    o2n = Cdb[Cdb['primary_representitive'] == True].set_index('subcluster')['primary_cluster'].to_dict()
    Cdb['primary_cluster'] = Cdb['subcluster'].map(o2n).astype(int)

    return Mdb, Cdb, cluster_ret

def cluster_mash_database(db, **kwargs):
    '''
    From a Mash database, cluster and return Cdb

    Args:
        db: Mdb (all_vs_all Mash results)

    Keyword arguments:
        clusterAlg: how to cluster database (default = single)
        P_ani: threshold to cluster at (default = 0.9)

    Returns:
        list: [Cdb, [linkage, linkage_db, arguments]]
    '''
    logging.debug('Clustering MASH database')

    # Load key words
    P_Lmethod = kwargs.get('clusterAlg','single')
    P_Lcutoff = 1 - kwargs.get('P_ani',.9)

    # Do the actual clustering
    db['dist'] = 1 - db['similarity']
    linkage_db = db.pivot("genome1","genome2","dist")
    Cdb, linkage = drep.d_cluster.cluster_utils.cluster_hierarchical(linkage_db, linkage_method= P_Lmethod, \
                                                                     linkage_cutoff= P_Lcutoff)
    Cdb = Cdb.rename(columns={'cluster':'primary_cluster'})
    Cdb['primary_cluster'] = Cdb['primary_cluster'].astype(int)

    # Preparing clustering for return
    arguments = {'linkage_method':P_Lmethod,'linkage_cutoff':P_Lcutoff,\
                    'comparison_algorithm':'MASH'}
    cluster_ret = [linkage, linkage_db, arguments]

    return Cdb, cluster_ret

def secondary_clustering(Bdb, Cdb, algorithm, data_folder, **kwargs):
    if kwargs.get('greedy_secondary_clustering', False) != True:
        Ndb = pd.DataFrame()
        for bdb, name in iteratre_clusters(Bdb, Cdb, id='primary_cluster'):
            logging.debug('running cluster {0}'.format(name))
            # logging.debug('total memory - {0:.2f} Mbp'.format(int(process.memory_info().rss)/1000000))
            ndb = compare_genomes(bdb, algorithm, data_folder, **kwargs)

            if len(ndb) == 0:
                logging.error("CRITICAL ERROR WITH PRIMARY CLUSTER {0}; TRYING AGAIN".format(name))
                ndb = compare_genomes(bdb, algorithm, data_folder, **kwargs)

            if len(ndb) > 0:
                ndb['primary_cluster'] = name
                Ndb = Ndb.append(ndb)
            else:
                logging.error("DOUBLE CRITICAL ERROR AGAIN WITH PRIMARY CLUSTER {0}; SKIPPING".format(name))

        # Run clustering on Ndb
        Cdb, c2ret = drep.d_cluster.utils._cluster_Ndb(Ndb, comp_method=algorithm, **kwargs)

        return Ndb, Cdb, c2ret

    else:
        return drep.d_cluster.greedy_clustering.greedy_secondary_clustering(Bdb, Cdb, algorithm, data_folder, **kwargs)


def iteratre_clusters(Bdb, Cdb, id='MASH_cluster'):
    Bdb = pd.merge(Bdb, Cdb)
    for cluster in Bdb[id].unique():
        d = Bdb[Bdb[id] == cluster]
        yield d, cluster

def compare_genomes(bdb, algorithm, data_folder, **kwargs):
    '''
    Compare a list of genomes using the algorithm specified

    This method takes in bdb (a table with the columns location and genome), runs
    pair-wise comparisons between all genomes in the sample, and returns a table
    with at least the columns 'reference', 'querry', 'ani','coverage', depending
    on what algorithm is called

    Args:
        bdb: DataFrame with ['genome', 'location'] (drep.d_filter.load_genomes)
        algorithm: options are ANImf, ANIn, gANI
        data_folder: location to store output files

    Keyword Arguments:
        wd: either this or prod_folder needed for gANI
        prod_folder: either this or wd needed for gANI

    Return:
        DataFrame: Ndb (['reference', 'querry', 'ani','coverage'])
    '''
    # To handle other versions of this method which passed in a WorkDirectory
    # instead of data_folder string
    if isinstance(data_folder, drep.WorkDirectory.WorkDirectory):
        data_folder = data_folder.get_dir('data')

    if not kwargs.get('greedy_secondary_clustering', False):
        if algorithm == 'ANImf':
            genome_list = bdb['location'].tolist()
            working_data_folder = os.path.join(data_folder, 'ANImf_files/')
            df = drep.d_cluster.external.run_pairwise_ANImf(genome_list, working_data_folder, **kwargs)
            return df

        elif algorithm == 'ANIn':
            genome_list = bdb['location'].tolist()
            working_data_folder = os.path.join(data_folder, 'ANIn_files/')
            df = drep.d_cluster.utils.run_pairwise_ANIn(genome_list, working_data_folder, **kwargs)
            return df

        elif algorithm == 'fastANI':
            genome_list = bdb['location'].tolist()
            working_data_folder = os.path.join(data_folder, 'fastANI_files/')
            df = drep.d_cluster.external.run_pairwise_fastANI(genome_list, working_data_folder, **kwargs)
            return df

        elif algorithm == 'gANI':
            # Figure out prodigal folder
            wd = kwargs.get('wd', False)
            if not wd:
                prod_folder = kwargs.pop('prod_folder', False)
                assert prod_folder != False
            else:
                prod_folder = wd.get_dir('prodigal')

            working_data_folder = os.path.join(data_folder, 'gANI_files/')
            df = drep.d_cluster.external.run_pairwise_gANI(bdb, working_data_folder, \
                                                           prod_folder=prod_folder, **kwargs)
            return df

        elif algorithm == 'goANI':
            # Figure out prodigal folder
            wd = kwargs.get('wd', False)
            if not wd:
                prod_folder = kwargs.pop('prod_folder', False)
                assert prod_folder != False
            else:
                prod_folder = wd.get_dir('prodigal')

            working_data_folder = os.path.join(data_folder, 'goANI_files/')
            df = drep.d_cluster.external.run_pairwise_goANI(bdb, working_data_folder, \
                                                            prod_folder=prod_folder, **kwargs)
            return df

        else:
            logging.error("{0} not supported".format(algorithm))
            sys.exit()

    else:
        SUPPORTED = ['fastANI']
        if algorithm not in SUPPORTED:
            message = f"{algorithm} is not supported for greedy secondary clustering!\nChoose one of the following supported S_algorithm options: {' '.join(SUPPORTED)}"
            logging.error(message)
            print(message)
            raise NameError

        working_data_folder = os.path.join(data_folder, 'greedy_clustering/')
        return drep.d_cluster.greedy_clustering.compare_genomes_greedy(bdb, algorithm, working_data_folder, **kwargs)





#!/usr/bin/env python3

from subprocess import call
import os
from Bio import SeqIO
import shutil
import multiprocessing
import multiprocessing.dummy
import datetime

import drep.d_filter
import drep.d_cluster
import drep.d_bonus

def run_cmd(cmd, dry=False, shell=True, logdir=False):
    # if dry, just print cmd
    if dry:
        if shell:
            print(cmd)
        else:
            print(' '.join(cmd))
        return

    # figure out what files you're going to store the output
    if logdir == False:
        devnull = open(os.devnull, 'w')
        sto = devnull
        ste = devnull
    else:
        uniq_filename = str(datetime.datetime.now().date()) + '_' + \
            str(datetime.datetime.now().time()).replace(':', '.')
        sto = open(os.path.join(logdir + uniq_filename + '.STDOUT'), 'w')
        ste = open(os.path.join(logdir + uniq_filename + '.STDERR'), 'w')

        # log the command
        now = open(os.path.join(logdir + uniq_filename + '.CMD'), 'w')
        if shell:
            now.write(str(cmd) + '\n')
        else:
            now.write(' '.join(cmd) + '\n')
        now.close()

    # run the command
    if shell:
        call(cmd,shell=True,stdout=sto, stderr=ste)
    else:
        call(cmd,stdout=sto, stderr=ste)
    return

def thread_cmd_wrapper(tup):
    run_cmd(*tup)

def thread_cmds(cmds, dry=False, shell=False, logdir=False, t=10):
    pool = multiprocessing.dummy.Pool(processes=t)
    tups = [(cmd, dry, shell, logdir) for cmd in cmds]
    pool.map(thread_cmd_wrapper, tups)
    pool.close()
    pool.join()
    return

def make_dir(outdirname,dry=False,overwrite=False):
    if dry:
        return
    if os.path.exists(outdirname):
        if overwrite:
            pass
            #shutil.rmtree(outdirname, ignore_errors=True)
            #os.makedirs(outdirname)
        else:
            raise ValueError("{0} already exsists! Will not overwrite with current settings".\
                            format(outdirname))
    else:
        os.makedirs(outdirname)

def clobber_dir(outdirname,dry=False,overwrite=False):
    if dry:
        return
    if os.path.exists(outdirname):
        if overwrite:
            shutil.rmtree(outdirname, ignore_errors=True)
            os.makedirs(outdirname)
        else:
            raise ValueError("{0} already exsists! Will not overwrite with current settings".\
                            format(outdirname))
    else:
        os.makedirs(outdirname)

def get_exe(name):
    '''
    Given the name of a program, return the path to the exe if I can find it

    Args:
        name: name of program

    Returns:
        string: location to program exe
    '''
    loc, works = drep.d_bonus.find_program(name)
    if works == False:
        raise ValueError("{0} isn't working- make sure its installed".format(name))
    return loc



#!/usr/bin/env python

import os
import sys
import gzip
import logging
import textwrap
import argparse

from Bio import SeqIO

def extract_bins(fasta, stb_file, out_base):
    if (out_base != '') & (out_base[-1] != '/'):
        out_base += '_'

    # Make scaffold to bin dictionary
    stb = {}
    stb_reader = open(stb_file)
    for line in stb_reader:
        line = line.strip()

        if line.startswith('#') or line.startswith('scaffold_name'):
            continue

        scaffold = line.split('\t')[0].strip()
        bin = line.split('\t')[1].strip()
        stb[scaffold] = bin

    # Do it with append
    if fasta[-3:] == '.gz':
        handle = gzip.open(fasta, "rt")
        seqs = SeqIO.parse(handle, "fasta")

    else:
        seqs = SeqIO.parse(fasta, "fasta")

    for seq_record in seqs:
        id = str(seq_record.id).strip()
        if id not in stb:
            # print("{0} not in stb".format(id))
            continue
        fasta = stb[id]

        nw = open("{0}{1}.fa".format(out_base, fasta), 'a')
        nw.write('\n'.join([">{0}".format(id), str(seq_record.seq), '']))
        nw.close()

    if fasta[-3:] == '.gz':
        handle.close()

def gen_stb(fastas):
    if ((len(fastas) == 1) & (not fastas[0].endswith('.gz'))):
        # See if this is a text file, not a fasta file
        text_list = True
        genomes = []
        with open(fastas[0], 'r') as o:
            for line in o.readlines():
                if line.startswith('>'):
                    text_list = False
                    break
                else:
                    genomes.append(line.strip())
        if text_list:
            print("Treating .fasta input as list")
            fastas = genomes

    stb = {}
    for fasta in fastas:
        bin = os.path.basename(fasta)
        if bin[-3:] == '.gz':
            with gzip.open(fasta, "rt") as handle:
                for seq_record in SeqIO.parse(handle, "fasta"):
                    id = str(seq_record.id).strip()
                    stb[id] = bin
        else:
            iter = SeqIO.parse(fasta, "fasta")
            for seq_record in iter:
                id = str(seq_record.id).strip()
                stb[id] = bin
    return stb

def print_stb(stb, output):
    file = open(output,'w')
    for key in sorted(stb, key=stb.get):
        file.write("{0}\t{1}\n".format(key,stb[key]))
    file.close()


class SmartFormatter(argparse.HelpFormatter):
    def _split_lines(self, text, width):
        if text.startswith('R|'):
            return text[2:].splitlines()
        # this is the RawTextHelpFormatter._split_lines
        return argparse.HelpFormatter._split_lines(self, text, width)

if __name__ == '__main__':
    parser = argparse.ArgumentParser(formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        usage=textwrap.dedent('''\
         \n
         The program has two uses related to scaffold to bin (.stb) files.
         .stb files should be tab-separated, with no header, and two columns: scaffold and bin

         Use 1) Pass a list of genomes to generate a .stb file.

         Example:
         parse_stb.py --reverse -f dereplicate_genomes/* -o representitve_genomes.stb

         Use 2) Pass a single .fasta file and a scaffold to bin file (.stb) to generate a number of
         fasta files based on the .stb file.

         Example:
         parse_stb.py -f concat_genomes.fasta -s scaffold_to_bin.tsv -o genomeList_1
         '''))

    parser.add_argument('-s','--stb',help='scaffold to bin file')
    parser.add_argument('-f','--fasta',help='fasta file to extract scaffolds from. Will treat as compressed if ends in .gz. This can also be a single text file with a genome on each line',nargs='*')
    parser.add_argument('-o','--output',help='output base name', default = '')

    parser.add_argument('--reverse',help='generate a stb from a list of genomes',\
                        action = "store_true")
    args = parser.parse_args()

    if args.reverse == False:
        if len(args.fasta) != 1:
            print("must give one and only one fasta file")
            sys.exit()
        extract_bins(args.fasta[0], args.stb, args.output)

    if args.reverse == True:
        stb = gen_stb(args.fasta)
        print_stb(stb, args.output)



#!/usr/bin/env python3
'''
d_analyze - a subset of drep

Make plots based on de-replication
'''
import matplotlib

import drep.d_cluster.cluster_utils
import drep.d_cluster.compare_utils
import drep.d_cluster.controller

matplotlib.use('Agg')

import logging
import math
import os
import sys

import pandas as pd
import seaborn as sns
import scipy.cluster.hierarchy
from sklearn import manifold
import numpy as np

from matplotlib import pyplot as plt
import matplotlib.ticker as ticker
from matplotlib.backends.backend_pdf import PdfPages
import matplotlib.patches as mpatches

import drep
import drep.d_cluster
import drep.d_filter

import traceback

import warnings
warnings.filterwarnings("ignore")#, category='all')

def d_analyze_wrapper(wd, **kwargs):
    '''
    Controller for the dRep analyze operation

    Args:
        wd: The current workDirectory
        **kwargs: Command line arguments

    Keyword Args:
        plots: List of plots to make [list of ints, 1-6]

    Returns:
        Makes some plots
    '''

    # Load the workDirectory
    wd = drep.WorkDirectory.WorkDirectory(wd)
    debug = kwargs.get('debug', False)

    # Figure out what plots to make
    options = ['1','2','3','4','5','6']

    to_plot = kwargs.get('plots', None)
    to_plot = _parse_plot_options(options, to_plot)
    logging.info("making plots {0}".format(', '.join(to_plot)))

    # Get the plot directory
    plot_dir = wd.get_dir('figures')

    # 1) Primary clustering dendrogram
    if '1' in to_plot:
        try:
            mash_dendrogram_from_wd(wd, plot_dir=plot_dir)
        except BaseException as e:
            logging.info('Failed to make plot #1: ' + str(e))
            if debug:
                traceback.print_exc()

    # 2) Secondary clustering dendrogram
    if '2' in to_plot:
        try:
            plot_secondary_dendrograms_from_wd(wd, plot_dir, **kwargs)
        except BaseException as e:
            logging.info('Failed to make plot #2: ' + str(e))
            if debug:
                traceback.print_exc()

    # 3) Secondary clusters MDS
    if '3' in to_plot:
        try:
            plot_secondary_mds_from_wd(wd, plot_dir, **kwargs)
        except BaseException as e:
            logging.info('Failed to make plot #3: ' + str(e))
            if debug:
                traceback.print_exc()

    # 4) Comparison scatterplots
    if '4' in to_plot:
        try:
            plot_scatterplots_from_wd(wd, plot_dir, **kwargs)
        except BaseException as e:
            logging.info('Failed to make plot #4: ' + str(e))
            if debug:
                traceback.print_exc()

    # 5) Complex bin scorring
    if '5' in to_plot:
        try:
            plot_binscoring_from_wd(wd, plot_dir, **kwargs)
        except BaseException as e:
            logging.info('Failed to make plot #5: ' + str(e))
            if debug:
                traceback.print_exc()

    # 6) Winning plot
    if '6' in to_plot:
        try:
            plot_winners_from_wd(wd, plot_dir, **kwargs)
        except BaseException as e:
            logging.info('Failed to make plot #6: ' + str(e))
            if debug:
                traceback.print_exc()


def mash_dendrogram_from_wd(wd, plot_dir=False):
    '''
    From the wd and kwargs, call plot_MASH_dendrogram

    Args:
        wd: WorkDirectory
        plot_dir (optional): Location to store figure

    Returns:
        Shows plot, makes a plot in the plot_dir
    '''
    # Load the required data
    try:
        Mdb = wd.get_db('Mdb', return_none=False, forPlotting=True)
        Cdb = wd.get_db('Cdb', return_none=False)
        Pcluster = wd.get_primary_linkage()
        Plinkage = Pcluster['linkage']
        clust_args = wd.arguments['cluster']
        PL_thresh = clust_args.get('P_ani', False)
        if PL_thresh != False:
            PL_thresh = 1-PL_thresh
    except:
        logging.error("Skipping plot 1 - you don't have all required dataframes")
        return

    if 'genome_chunk' in list(Mdb.columns):
        logging.error("Skipping plot 1 - cannot generate with multiround_primary_clustering enabled")
        return

    # Make the plot
    logging.info("Plotting primary dendrogram")
    plot_MASH_dendrogram(Mdb, Cdb, Plinkage, threshold = PL_thresh,\
                    plot_dir = plot_dir)

def plot_secondary_dendrograms_from_wd(wd, plot_dir, **kwargs):
    '''
    From the wd and kwargs, make the secondary dendrograms

    Args:
        wd: WorkDirectory
        plot_dir (optional): Location to store figure

    Returns:
        Makes plot
    '''

    # Load required databases
    try:
        Ndb = wd.get_db('Ndb', return_none=False)
        Cdb = wd.get_db('Cdb', return_none=False)
    except:
        logging.error("Skipping plot 2 - you don't have all required dataframes")
        return

    if len(Cdb[Cdb['cluster_method'] == 'greedy']) > 0:
        logging.error("Skipping plot 2 - cannot generate with greedy_secondary_clustering enabled")
        return

    # Initialize a .pdf
    if plot_dir != False:
        pp = PdfPages(plot_dir + 'Secondary_clustering_dendrograms.pdf')
        save = True
    else:
        save = False

    logging.info("Plotting secondary dendrograms")

    # Load winner database if it exists
    if wd.hasDb('Wdb'):
        Wdb = wd.get_db('Wdb')
        winners = Wdb['genome'].unique()
        kwargs['winners'] = winners

    # Load genome 2 taxonomy if it exists
    genome2taxonomy = _get_genome2taxonomy(wd)
    kwargs['genome2taxonomy'] = genome2taxonomy

    # For every cluster:
    for cluster in sorted(Cdb['primary_cluster'].unique()):
        d = Cdb[Cdb['primary_cluster'] == cluster]

        # Skip if it's a singleton
        if len(d['genome'].unique()) == 1:
            continue

        # Load the linkage information
        linkI = wd.get_cluster("secondary_linkage_cluster_{0}".format(cluster))
        db = linkI['db']
        linkage = linkI['linkage']
        args = linkI['arguments']
        threshold = args['linkage_cutoff']
        alg = args['comparison_algorithm']
        clust_alg = args['linkage_method']
        min_cov = args['minimum_coverage']

        kwargs['threshold'] = threshold
        kwargs['title_string'] = 'Primary cluster {0}'.format(cluster)
        kwargs['subtitle_string'] = "Comp method: {0}    ".format(alg) +\
                "Clust method: {0}    Min cov: {1}".format(clust_alg, min_cov)

        # Get name2cluster
        names = list(db.columns)
        name2cluster = Cdb.set_index('genome')['secondary_cluster'].to_dict()
        for name in names: # Handle the case where you deleted a secondary cluster
            if name not in Cdb['genome'].tolist():
                name2cluster[name] = '0_0'
        kwargs['name2cluster'] = name2cluster

        kwargs['self_thresh'] = get_highest_self(Ndb, names)

        # Make the dendrogram
        _make_special_dendrogram(linkage, names, **kwargs)

        # Save the file
        fig = plt.gcf()
        if save == True:
            pp.savefig(fig)
        plt.show()
        plt.close(fig)

    pp.close()
    plt.close('all')

def plot_secondary_mds_from_wd(wd, plot_dir, **kwargs):
    '''
    Make a .pdf of MDS of each cluster

    Args:
        wd: WorkDirectory
        plot_dir (optional): Location to store figure

    Returns:
        Makes plot
    '''
    # Load required databases
    try:
        Ndb = wd.get_db('Ndb', return_none=False)
        Cdb = wd.get_db('Cdb', return_none=False)
    except:
        logging.error("Skipping plot 3 - you don't have all required dataframes")
        return

    logging.info("Plotting MDS plot")
    # initialize a .pdf
    if plot_dir != False:
        pp = PdfPages(plot_dir + 'Secondary_clustering_MDS.pdf')
        save = True
    else:
        save = False

    # for every cluster:
    Cdb = wd.get_db('Cdb')
    for cluster in sorted(Cdb['primary_cluster'].unique()):
        d = Cdb[Cdb['primary_cluster'] == cluster]

        # Skip if it's a singleton
        if len(d['genome'].unique()) == 1:
            continue

        # Load the linkage information
        linkI = wd.get_cluster("secondary_linkage_cluster_{0}".format(cluster))
        db = linkI['db']
        args = linkI['arguments']

        # Load name to cluster
        names = list(db.columns)
        name2cluster = Cdb.set_index('genome')['secondary_cluster'].to_dict()
        for name in names: # Handle the case where you deleted a secondary cluster
            if name not in Cdb['genome'].tolist():
                name2cluster[name] = '0_0'

        # Get the colors
        name2color = gen_color_dictionary(names, name2cluster)
        colors = [name2color[n] for n in names]

        # Make cluster 2 color
        cluster2color = {name2cluster[n]: name2color[n] for n in names}

        # make the mds plot
        _make_mds_plot(cluster, db, names, colors=colors, annotate=False,
                      cluster2color = cluster2color)

        # save the plot
        fig = plt.gcf()
        if save == True:
            pp.savefig(fig, bbox_inches='tight')
        plt.show()
        plt.close(fig)

    pp.close()
    plt.close('all')

def plot_scatterplots_from_wd(wd, plot_dir, **kwargs):
    '''
    From the wd and kwargs, call plot_scatterplots

    Args:
        wd: WorkDirectory
        plot_dir (optional): Location to store figure

    Returns:
        Shows plot, makes a plot in the plot_dir
    '''
    # Load the required data
    try:
        Ndb = wd.get_db('Ndb', return_none=False)
        Mdb = wd.get_db('Mdb', return_none=False)
        Cdb = wd.get_db('Cdb', return_none=False)
    except:
        logging.error("Skipping plot 4 - you don't have all required dataframes")
        return

    # Make the plot
    logging.info("Plotting scatterplots")
    plot_scatterplots(Mdb, Ndb, Cdb, plot_dir = plot_dir)

def plot_binscoring_from_wd(wd, plot_dir, **kwargs):
    '''
    From the wd and kwargs, call plot_winner_scoring_complex

    Args:
        wd: WorkDirectory
        plot_dir (optional): Location to store figure

    Returns:
        Shows plot, makes a plot in the plot_dir
    '''
    # Load the required data
    try:
        Sdb = wd.get_db('Sdb', return_none=False)
        Cdb = wd.get_db('Cdb', return_none=False)
        Wdb = wd.get_db('Wdb', return_none=False)
        Bdb = wd.get_db('Bdb', return_none=False)
    except:
        logging.error("Skipping plot 5 - you don't have all required dataframes")
        return

    # Deal with genome quality
    Gdb = drep.d_filter._get_run_genomeInfo(wd, Bdb, no_run=True)

    # Only keep things you want
    Gdb = Gdb[[c for c in Gdb.columns if c in ['genome', 'location', 'N50', 'length', 'completeness', 'contamination', 'strain_heterogeneity']]]

    # Make the plot
    logging.info("Plotting bin scorring plot")
    plot_winner_scoring_complex(Wdb, Sdb, Cdb, Gdb, plot_dir = plot_dir, **kwargs)

def plot_winners_from_wd(wd, plot_dir, **kwargs):
    '''
    From the wd and kwargs, call plot_winners

    Args:
        wd: WorkDirectory
        plot_dir: Location to store figure

    Returns:
        Shows plot, makes a plot in the plot_dir
    '''
    # Load the required data
    try:
        Wdb = wd.get_db('Wdb', return_none=False)
        Bdb = wd.get_db('Bdb', return_none=False)
    except:
        logging.error("Skipping plot 6 - you don't have all required dataframes")
        return

    # Deal with genome quality
    Gdb = drep.d_filter._get_run_genomeInfo(wd, Bdb, no_run=True)

    # Only keep things you want
    Gdb = Gdb[[c for c in Gdb.columns if c in ['genome', 'location', 'N50', 'length', 'completeness', 'contamination', 'strain_heterogeneity']]]

    # Get optional data
    Wndb = wd.get_db('Wndb')
    Wmdb = wd.get_db('Wmdb')
    Widb = wd.get_db('Widb')

    # Make the plot
    logging.info("Plotting winning genomes plot...")
    plot_winners(Wdb, Gdb, Wndb, Wmdb, Widb, plot_dir = plot_dir, **kwargs)

def plot_scatterplots(Mdb, Ndb, Cdb, plot_dir=False):
    '''
    Make scatterplots comparing genome comparison algorithms

    *   plot_MASH_vs_ANIn_ani(Mdb, Ndb)
        - Plot MASH_ani vs. ANIn_ani (including correlation)

    *   plot_MASH_vs_ANIn_cov(Mdb, Ndb)
        - Plot MASH_ani vs. ANIn_cov (including correlation)

    *   plot_ANIn_vs_ANIn_cov(Mdb, Ndb)
        - Plot ANIn vs. ANIn_cov (including correlation)

    *   plot_MASH_vs_len(Mdb, Ndb)
        - Plot MASH_ani vs. length_difference (including correlation)

    *   plot_ANIn_vs_len(Ndb)
        - Plot ANIn vs. length_difference (including correlation)

    Args:
        Mdb: DataFrame of Mash comparison results
        Ndb: DataFrame of secondary clustering results
        Cdb: DataFrame of Clustering results
        plot_dir (optional): Location to store plot

    Return:
        Makes and shows plot
    '''
    sns.set_style('whitegrid')

    # Initialize a .pdf
    if plot_dir != False:
        pp = PdfPages(plot_dir + 'Clustering_scatterplots.pdf')
        save = True
    else:
        save = False

    g = plot_MASH_vs_ANIn_ani(Mdb,Ndb,exclude_zero_MASH=False)
    if save: pp.savefig(g)
    plt.show()

    g = plot_MASH_vs_secondary_ani(Mdb,Ndb,Cdb,exclude_zero_MASH=False)
    if save: pp.savefig(g)
    plt.show()

    g = plot_MASH_vs_ANIn_cov(Mdb,Ndb)
    if save: pp.savefig(g)
    plt.show()

    g = plot_ANIn_vs_ANIn_cov(Ndb)
    if save: pp.savefig(g)
    plt.show()

    try:
        g = plot_MASH_vs_len(Mdb,Ndb)
        if save: pp.savefig(g)
        plt.show()
    except:
        pass

    try:
        g = plot_ANIn_vs_len(Mdb,Ndb)
        if save: pp.savefig(g)
        plt.show()
    except:
        pass

    if save:
        pp.close()
    plt.close('all')

def plot_MASH_vs_ANIn_ani(Mdb, Ndb, exclude_zero_MASH=True):
    '''
    Makes plot and retuns plt.cgf()

    All parameters are obvious
    '''
    plt.close('all')
    mdb = Mdb.copy()
    mdb.rename(columns={'genome1':'querry','genome2':'reference',
                        'similarity':'MASH_ANI'},inplace=True)
    if exclude_zero_MASH:
        mdb= mdb[mdb['MASH_ANI'] > 0]

    db = pd.merge(mdb,Ndb)
    db.rename(columns={'ani':'ANIn'},inplace=True)
    g = sns.jointplot(x='ANIn',y='MASH_ANI',data=db)
    plt.gcf().suptitle('MASH vs ANIn comparisons (all)')
    plt.subplots_adjust(top=0.9)
    #plt.subplots_adjust(left=0.2)
    return plt.gcf()

def plot_MASH_vs_secondary_ani(Mdb,Ndb,Cdb,exclude_zero_MASH=True):
    '''
    Makes plot and retuns plt.cgf()

    All parameters are obvious
    '''
    plt.close('all')
    Xdb = pd.DataFrame()

    # Make a database of all comparisons
    mdb = Mdb.copy()
    mdb.rename(columns={'genome1':'querry','genome2':'reference',
                        'similarity':'MASH_ANI'},inplace=True)
    if exclude_zero_MASH:
        mdb= mdb[mdb['MASH_ANI'] > 0]
    db = pd.merge(mdb,Ndb)
    db.rename(columns={'ani':'ANIn'},inplace=True)

    # Filter to only comparisons within secondary clusters
    g2c = Cdb.set_index('genome')['secondary_cluster'].to_dict()
    db['ref_secondary_cluster'] = db['reference'].map(g2c)
    db['qu_secondary_cluster'] = db['querry'].map(g2c)
    db = db[db['ref_secondary_cluster'] == db['qu_secondary_cluster']]

    g = sns.jointplot(x='ANIn',y='MASH_ANI',data=db)
    plt.gcf().suptitle('MASH vs ANIn comparisons (within secondary clusters only)')
    plt.subplots_adjust(top=0.9)
    #plt.subplots_adjust(left=0.2)
    return plt.gcf()

def plot_MASH_vs_ANIn_cov(Mdb,Ndb,exclude_zero_MASH=True):
    '''
    Makes plot and retuns plt.cgf()

    All parameters are obvious
    '''
    plt.close('all')
    mdb = Mdb.copy()
    mdb.rename(columns={'genome1':'querry','genome2':'reference',
                        'similarity':'MASH_ANI'},inplace=True)
    if exclude_zero_MASH:
        mdb= mdb[mdb['MASH_ANI'] > 0]

    db = pd.merge(mdb,Ndb)
    db.rename(columns={'alignment_coverage':'ANIn_alignment_coverage'},inplace=True)
    g = sns.jointplot(x='ANIn_alignment_coverage',y='MASH_ANI',data=db)
    return plt.gcf()

def plot_ANIn_vs_ANIn_cov(Ndb):
    '''
    Makes plot and retuns plt.cgf()

    All parameters are obvious
    '''
    plt.close('all')
    db = Ndb.copy()
    db.rename(columns={'alignment_coverage':'ANIn_alignment_coverage','ani':'ANIn'},inplace=True)
    g = sns.jointplot(y='ANIn',x='ANIn_alignment_coverage',data=db)
    return plt.gcf()

def plot_MASH_vs_len(Mdb,Ndb,exclude_zero_MASH=True):
    '''
    Makes plot and retuns plt.cgf()

    All parameters are obvious
    '''
    plt.close('all')
    mdb = Mdb.copy()
    mdb.rename(columns={'genome1':'querry','genome2':'reference',
                        'similarity':'MASH_ANI'},inplace=True)
    if exclude_zero_MASH:
        mdb= mdb[mdb['MASH_ANI'] > 0]

    db = pd.merge(mdb,Ndb)
    db['length_difference'] = abs(db['reference_length'] - db['querry_length'])
    g = sns.jointplot(x='MASH_ANI',y='length_difference',data=db)

    # Make a decending xaxis
    axs = plt.gcf().get_axes()
    plt.sca(axs[0])
    plt.xlim(db['MASH_ANI'].max(),db['MASH_ANI'].min())

    plt.gcf().suptitle('MASH vs length difference of genomes compared')
    plt.subplots_adjust(top=0.9)
    plt.subplots_adjust(left=0.2)

    return plt.gcf()

def plot_ANIn_vs_len(Mdb,Ndb,exclude_zero_MASH=True):
    '''
    Makes plot and retuns plt.cgf()

    All parameters are obvious
    '''
    plt.close('all')
    mdb = Mdb.copy()
    mdb.rename(columns={'genome1':'querry','genome2':'reference',
                        'similarity':'MASH_ANI'},inplace=True)
    if exclude_zero_MASH:
        mdb= mdb[mdb['MASH_ANI'] > 0]

    db = pd.merge(mdb,Ndb)
    db['length_difference'] = abs(db['reference_length'] - db['querry_length'])
    db.rename(columns={'ani':'ANIn'},inplace=True)
    g = sns.jointplot(x='ANIn',y='length_difference',data=db)

    # Make a decending xaxis
    axs = plt.gcf().get_axes()
    plt.sca(axs[0])
    plt.xlim(db['ANIn'].max(),db['ANIn'].min())

    plt.gcf().suptitle('ANIn vs length difference of genomes compared')
    plt.subplots_adjust(top=0.9)
    plt.subplots_adjust(left=0.2)
    return plt.gcf()

"""
CLUSETER PLOTS
"""

def plot_MASH_dendrogram(Mdb, Cdb, linkage, threshold=False, plot_dir=False):
    '''
    Make a dendrogram of the primary clustering

    Args:
        Mdb: DataFrame of Mash comparison results; make sure loaded not as categories
        Cdb: DataFrame of Clustering results
        linkage: Result of scipy.cluster.hierarchy.linkage
        threshold (optional): Line to plot on x-axis
        plot_dir (optional): Location to store plot

    Returns:
        Makes and shows plot
    '''
    sns.set_style('white',{'axes.grid': False})

    if Mdb['genome1'].dtype.name == 'category':
        logging.error("WARNING: Primary dendrogram labels may be shuffled! Load as csv to prevent this")

    db = Mdb.pivot("genome1","genome2","similarity")
    names = list(db.columns)
    name2cluster = Cdb.set_index('genome')['primary_cluster'].to_dict()
    name2color = gen_color_dictionary(names, name2cluster)

    # Make the dendrogram
    g = fancy_dendrogram(linkage,names,name2color,threshold=threshold)
    plt.title('MASH clustering')
    plt.xlabel('MASH Average Nucleotide Identity (ANI)')

    #plt.xlim([0,.4])

    sns.despine(left=True,top=True,right=True,bottom=False)

    # Adjust the figure size
    fig = plt.gcf()
    fig.set_size_inches(10,_x_fig_size(len(names),factor=.2))
    plt.subplots_adjust(left=0.3)

    # Adjust the x labels
    plt.tick_params(axis='both', which='major', labelsize=8)
    axes = plt.gca()
    labels = axes.xaxis.get_majorticklocs()
    for i, label in enumerate(labels):
        labels[i] = float("{0:.2f}".format((1 - float(label)) * 100))
    axes.set_xticklabels(labels)

    # Add cluster to the y axis
    g2c = Cdb.set_index('genome')['secondary_cluster'].to_dict()
    axes = plt.gca()
    labels = [item.get_text() for item in axes.get_yticklabels()]
    for i, label in enumerate(labels):
        labels[i] = "{0} ({1})".format(label, g2c[label])
    axes.set_yticklabels(labels)

    # Save the figure
    if plot_dir != None:
        plt.savefig(os.path.join(plot_dir, 'Primary_clustering_dendrogram.pdf'),\
            format="pdf", transparent=True, bbox_inches='tight')
    plt.show()
    plt.close('all')

"""
WINNER PLOTS
"""

def plot_winner_scoring_complex(Wdb, Sdb, Cdb, Gdb, plot_dir= False, **kwargs):
    '''
    Make a plot showing the genome scoring for all genomes

    Args:
        Wdb: DataFrame of winning dereplicated genomes
        Sdb: Scores of all genomes
        Cdb: DataFrame of Clustering results
        Gdb: DataFrame of genome scoring information
        plot_dir (optional): Location to store plot

    Returns:
        makes plot
    '''
    # Set style
    sns.reset_orig()

    # Initialize a .pdf
    if plot_dir != False:
        pp = PdfPages(os.path.join(plot_dir, 'Cluster_scoring.pdf'))
        save = True
    else:
        save = False

    # Figure out what you're going to show
    bars = _get_toshow(Gdb)
    bars += ['score']

    # Get winners
    winners = list(Wdb['genome'].unique())

    for cluster in sorted(Cdb['secondary_cluster'].unique(), key=lambda x: _comp_cluster(x)):
        # Make a db for this cluster
        d = Cdb[Cdb['secondary_cluster'] == cluster]
        d = d.merge(Sdb, how='left', on= 'genome')
        d = d.merge(Gdb, how='left', on= 'genome')
        d = d[bars + ['genome']]

        # Make the normalize bar plot
        nd = normalize(d)
        db = pd.melt(nd, id_vars=['genome'], value_vars=bars)
        g = sns.barplot(data=db, y='genome', x='value', hue='variable')

        # Get a list of the un-normalized values
        x = pd.melt(d, id_vars=['genome'], value_vars=bars)
        vals = []
        for variable in x['variable'].unique():
            vals += [v for v in x['value'][x['variable'] == variable].tolist()]

        # Add un-normalized values to barplots
        i = 0
        for p in g.patches:
            g.annotate("{0:.1f}".format(vals[i]), (p.get_width(), p.get_y()+(p.get_height()/1.1) ), fontsize=8)
            i += 1

        plt.title('Scoring of cluster {0}'.format(cluster))
        plt.xlabel('Normalized Score')
        plt.legend(loc=(0,0), fancybox=True, framealpha=0.5)
        plt.tick_params(axis='both', which='major', labelsize=8)

        # Mark winning one
        labels = d['genome'].tolist()
        for i, label in enumerate(labels):
            if label in winners: labels[i] = label + ' *'
        axes = plt.gca()
        axes.set_yticklabels(labels)

        # Add taxonomy
        if kwargs.get('genome2taxonomy',False) != False:
            g2t = kwargs.get('genome2taxonomy')
            axes = plt.gca()
            labels = [item.get_text() for item in axes.get_yticklabels()]
            for i, label in enumerate(labels):
                labels[i] = "{0}\n{1}".format(label, g2t[label.replace(' *','')])
            axes.set_yticklabels(labels)

        fig = plt.gcf()
        fig.set_size_inches(12,_x_fig_size(len(labels), factor=1))
        plt.subplots_adjust(left=0.5)

        if save == True:
            pp.savefig(fig)
        plt.show()
        plt.close(fig)

    if save:
        pp.close()
    plt.close('all')

def plot_winners(Wdb, Gdb, Wndb, Wmdb, Widb, plot_dir= False, **kwargs):
    '''
    Make a bunch of plots about the de-replicated genomes

    THIS REALLY NEEDS IMPROVED UPON
    '''

    # Set style
    sns.reset_orig()

    # Initialize a .pdf
    if plot_dir != False:
        pp = PdfPages(plot_dir + 'Winning_genomes.pdf')
        save = True
    else:
        save = False

    # Make piecharts
    if Widb is not None:
        labels = []
        sizes = []
        for com in Widb['completeness_metric'].unique():
            d = Widb[Widb['completeness_metric'] == com]
            labels.append(com)
            sizes.append(len(d['genome'].unique()))
        labels = _annotate_labels(labels,'comp')

        if (len(labels) != len(sizes)) | (len(labels) == 0): # not sure when this would happen, but it does...
            logging.debug("len(labels) != len(sizes); {0} vs {1}".format(\
                len(labels), len(sizes)))

        else:
            _make_piechart(labels,sizes)
            plt.title('Overall Winner Completeness')

            # Save this page
            if save == True:
                fig = plt.gcf()
                pp.savefig(fig)
            plt.show()
            plt.close(fig)

            labels = []
            sizes = []
            for com in Widb['contamination_metric'].unique():
                d = Widb[Widb['contamination_metric'] == com]
                labels.append(com)
                sizes.append(len(d['genome'].unique()))
            labels = _annotate_labels(labels,'con')
            _make_piechart(labels,sizes)
            plt.title('Overall Winner Contamination')

            # Save this page
            if save == True:
                fig = plt.gcf()
                pp.savefig(fig)
            plt.show()
            plt.close(fig)

    # Figure out what you're going to show
    bars = _get_toshow(Gdb)
    bars += ['score']

    # Make a db for the winners
    d = Wdb.sort_values('score', ascending=False)
    d = d.merge(Gdb, how='left', on= 'genome')
    d = d[bars + ['genome']]

    # Make the scoring plot
    _make_scoring_plot(d,bars,**kwargs)
    plt.title('Scoring of winning genomes')

    # Save this page
    if save == True:
        fig = plt.gcf()
        pp.savefig(fig)
    plt.show()
    plt.close(fig)

    if Wmdb is not None:
        # Make a MASH linkage for the dendrogram
        db = Wmdb.copy()
        db['dist'] = 1 - db['similarity']
        linkage_db = db.pivot("genome1","genome2","dist")
        names = list(linkage_db.columns)
        Cdb, linkage = drep.d_cluster.cluster_utils.cluster_hierarchical(linkage_db, linkage_method='average', \
                                                                         linkage_cutoff= 0)

        # Make the MASH dendrogram
        _make_dendrogram(linkage,names)
        plt.title('MASH dendrogram')

        # Save this page
        if save == True:
            fig = plt.gcf()
            pp.savefig(fig)
        plt.show()
        plt.close(fig)

    if Wndb is not None:
        # Make a ANIn linkage for the dendrogram
        d = Wndb.copy()
        drep.d_cluster.add_avani(d)
        #d['av_ani'] = d.apply(lambda row: drep.d_cluster.average_ani (row,d),axis=1)
        d['dist'] = 1 - d['av_ani']
        db = d.pivot("reference", "querry", "dist")
        names = list(db.columns)
        Cdb, linkage = drep.d_cluster.cluster_utils.cluster_hierarchical(db, linkage_method='average', \
                                                                         linkage_cutoff= 0)

        # Make the ANIn dendrogram
        _make_dendrogram(linkage,names)
        plt.title('ANIn dendrogram (NOT filtered for alignment length)')

        # Save this page
        if save == True:
            fig = plt.gcf()
            pp.savefig(fig)
        plt.show()
        plt.close(fig)

        # Make a ANIn linkage for the filtered dendrogram
        d = Wndb.copy()
        d.loc[d['alignment_coverage'] <= 0.1, 'ani'] = 0
        drep.d_cluster.add_avani(d)
        #d['av_ani'] = d.apply(lambda row: drep.d_cluster.average_ani (row,d),axis=1)
        d['dist'] = 1 - d['av_ani']
        db = d.pivot("reference", "querry", "dist")
        names = list(db.columns)
        Cdb, linkage = drep.d_cluster.cluster_utils.cluster_hierarchical(db, linkage_method='average', \
                                                                         linkage_cutoff= 0)

        # Make the ANIn dendrogram
        _make_dendrogram(linkage,names)
        plt.title('ANIn dendrogram (filtered for 10% alignment)')

        # Save this page
        if save == True:
            fig = plt.gcf()
            pp.savefig(fig)
        plt.show()
        plt.close(fig)

    # Save the .pdf
    if save:
        pp.close()
    plt.close('all')

def calc_dist(x1, y1, x2, y2):
    '''
    Return distance from two points

    Args: self explainatory

    Returns:
        int: distance
    '''
    dist = math.hypot(x2 - x1, y2 - y1)
    return dist

def get_highest_self(db, genomes, min = 1.0e-4):
    '''
    Return the highest ANI value resulting from comparing a genome to itself
    '''
    d = db[db['reference'].isin(genomes)]
    self_thresh = 1 - d['ani'][d['reference'] == d['querry']].min()

    # Because 0s don't show up on the graph
    if self_thresh == float(0):
        self_thresh =  min
    return self_thresh

def _make_piechart(labels,sizes):
    '''
    Used by winner plot
    '''
    plt.pie(sizes,labels=labels,startangle=45,\
            autopct=_make_autopct(sizes),\
            shadow = True)
    plt.axis('equal')

def _make_autopct(values):
    '''
    Used by winner plot
    '''
    def my_autopct(pct):
        total = sum(values)
        val = int(round(pct*total/100.0))
        return '{p:.2f}%  ({v:d})'.format(p=pct,v=val)
    return my_autopct

def _annotate_labels(labels,how):
    '''
    Used by winner plot
    '''
    if how == 'comp':
        labs = []
        for label in labels:
            if label == 'near':
                labs.append('near (>90%)')
            if label == 'perfect':
                labs.append('perfect (100%)')
            if label == 'substantial':
                labs.append('substantial (>70%)')
            if label == 'moderate':
                labs.append('moderate (>50%)')
            if label == 'partial':
                labs.append('partial (<50%)')
        return labs

    if how == 'con':
        labs = []
        for label in labels:
            if label == 'low':
                labs.append('low (<5%)')
            if label == 'none':
                labs.append('none (0%)')
            if label == 'medium':
                labs.append('medium (<10%)')
            if label == 'high':
                labs.append('high (<15%)')
            if label == 'very high':
                labs.append('very high (>15%)')
        return labs

def _x_fig_size(points, factor= .07, min= 8):
    '''
    Calculate how big the x of the figure should be
    '''
    size = points * factor
    return max([size,min])

def fancy_dendrogram(linkage,names,name2color=False,threshold=False,self_thresh=False):
    '''
    Make a fancy dendrogram
    '''
    # Make the dendrogram
    if threshold == False:
        scipy.cluster.hierarchy.dendrogram(linkage,labels=names,orientation='right')
    else:
        scipy.cluster.hierarchy.dendrogram(linkage,labels=names, color_threshold=threshold,\
                                            orientation='right')

    # Color the names
    if name2color != False:
        ax = plt.gca()
        xlbls = ax.get_ymajorticklabels()
        for lbl in xlbls:
            lbl.set_color(name2color[lbl.get_text()])

    # Add the threshold
    if threshold:
        plt.axvline(x=threshold, c='k', linestyle='dotted')
    if self_thresh != False:
        plt.axvline(x=self_thresh, c='red', linestyle='dotted', lw=1)

    g = plt.gcf()
    return g

def normalize(df):
    '''
    Normalize all columns in df to 0-1 except 'genome' or 'location'

    Args:
        df: DataFrame

    Return:
        DataFrame: Nomralized
    '''
    result = df.copy()
    for feature_name in df.columns:
        if feature_name in ['genome', 'location']:
            continue
        if not np.issubdtype(df[feature_name].dtype, np.number):
            continue
        max_value = max(df[feature_name].max(),0)
        result[feature_name] = [max((x / max_value) if max_value != 0 else 0,0) for x in result[feature_name].tolist()]

    return result

def gen_color_list(names,name2cluster):
    '''
    Make a list of colors the same length as names, based on their cluster
    '''
    cm = plt.get_cmap('gist_rainbow')

    # 1. generate cluster to color
    cluster2color = {}
    clusters = set(name2cluster.values())
    NUM_COLORS = len(clusters)
    for cluster in clusters:
        try:
            cluster2color[cluster] = cm(1.*int(cluster)/NUM_COLORS)
        except:
            cluster2color[cluster] = cm(1.*int(str(cluster).split('_')[1])/NUM_COLORS)

    #2. generate list of colors
    colors = []
    for name in names:
        colors.append(cluster2color[name2cluster[name]])

    return colors

def gen_color_dictionary(names, name2cluster):
    '''
    Make the dictionary name2color

    Args:
        names: key in the returned dictionary
        name2cluster: a dictionary of name to it's cluster

    Returns:
        dict: name -> color
    '''
    #cm = _rand_cmap(len(set(name2cluster.values()))+1,type='bright')
    vals = np.linspace(0,1,len(set(name2cluster.values()))+1)
    np.random.shuffle(vals)
    cm = plt.cm.colors.ListedColormap(plt.cm.jet(vals))

    # 1. generate cluster to color
    cluster2color = {}
    clusters = set(name2cluster.values())
    NUM_COLORS = len(clusters)
    for cluster in clusters:
        try:
            cluster2color[cluster] = cm(1.*int(cluster)/NUM_COLORS)
        except:
            cluster2color[cluster] = cm(1.*int(str(cluster).split('_')[1])/NUM_COLORS)

    #2. name to color
    name2color = {}
    for name in names:
        name2color[name] = cluster2color[name2cluster[name]]

    return name2color

def _comp_cluster(c):
    '''
    Used in secondary dendrogram creation
    '''
    first = int(c.split('_')[0])
    dec = float(c.split('_')[1])
    return first + dec/100

def _rand_cmap(nlabels, type='bright', first_color_black=True, last_color_black=False, verbose=False):
    """
    Creates a random colormap to be used together with matplotlib. Useful for segmentation tasks
    :param nlabels: Number of labels (size of colormap)
    :param type: 'bright' for strong colors, 'soft' for pastel colors
    :param first_color_black: Option to use first color as black, True or False
    :param last_color_black: Option to use last color as black, True or False
    :param verbose: Prints the number of labels and shows the colormap. True or False
    :return: colormap for matplotlib
    """
    from matplotlib.colors import LinearSegmentedColormap
    import colorsys
    import numpy as np


    if type not in ('bright', 'soft'):
        print ('Please choose "bright" or "soft" for type')
        return

    if verbose:
        print('Number of labels: ' + str(nlabels))

    # Generate color map for bright colors, based on hsv
    if type == 'bright':
        randHSVcolors = [(np.random.uniform(low=0.0, high=1),
                          np.random.uniform(low=0.2, high=1),
                          np.random.uniform(low=0.9, high=1)) for i in range(nlabels)]

        # Convert HSV list to RGB
        randRGBcolors = []
        for HSVcolor in randHSVcolors:
            randRGBcolors.append(colorsys.hsv_to_rgb(HSVcolor[0], HSVcolor[1], HSVcolor[2]))

        if first_color_black:
            randRGBcolors[0] = [0, 0, 0]

        if last_color_black:
            randRGBcolors[-1] = [0, 0, 0]

        random_colormap = LinearSegmentedColormap.from_list('new_map', randRGBcolors, N=nlabels)

    # Generate soft pastel colors, by limiting the RGB spectrum
    if type == 'soft':
        low = 0.6
        high = 0.95
        randRGBcolors = [(np.random.uniform(low=low, high=high),
                          np.random.uniform(low=low, high=high),
                          np.random.uniform(low=low, high=high)) for i in range(nlabels)]

        if first_color_black:
            randRGBcolors[0] = [0, 0, 0]

        if last_color_black:
            randRGBcolors[-1] = [0, 0, 0]
        random_colormap = LinearSegmentedColormap.from_list('new_map', randRGBcolors, N=nlabels)

    # Display colorbar
    if verbose:
        from matplotlib import colors, colorbar
        from matplotlib import pyplot as plt
        fig, ax = plt.subplots(1, 1, figsize=(15, 0.5))

        bounds = np.linspace(0, nlabels, nlabels + 1)
        norm = colors.BoundaryNorm(bounds, nlabels)

        cb = colorbar.ColorbarBase(ax, cmap=random_colormap, norm=norm, spacing='proportional', ticks=None,
                                   boundaries=bounds, format='%1i', orientation=u'horizontal')

    return random_colormap


def _parse_plot_options(options, args):
    '''
    Read user input and figure out a list of plots to make

    Args:
        options: list of possible plots to make (default [1-6])
        args: the command line passed in

    Returns:
        list: list of ints in the args
    '''
    to_plot = []

    if args == []:
        return []

    if args[0] in ['all','a']:
        to_plot += options

    elif args == None:
        logging.error("No plots given!")
        sys.exit()
        return None

    else:
        for arg in args:
            if arg in options:
                to_plot.append(arg)
            else:
                for letter in arg:
                    if letter in options:
                        to_plot.append(letter)
                    else:
                        logging.error("Can't interpret plotting argument {0}! quitting".format(arg))
                        sys.exit()
    return to_plot

def _get_genome2taxonomy(wd):
    '''
    Return dictionary: genome -> taxonomy

    All based on Bdb at the moment

    Return False if can't do it
    '''
    try:
        Bdb = wd.get_db('Bdb')
        if 'taxonomy' in Bdb:
            genome2taxonomy = Bdb.set_index('genome')['taxonomy'].to_dict()
        return genome2taxonomy
    except:
        return False

def _make_dendrogram(linkage, names, **kwargs):
    '''
    Currently used by winner plot

    names can be gotten like:
    db = db.pivot("reference","querry","ani")
    names = list(db.columns)
    '''
    threshold = kwargs.get('threshold',False)
    title = kwargs.get('title',False)

    # Make the dendrogram
    g = fancy_dendrogram(linkage,names,threshold=threshold)
    plt.title(title)
    plt.xlabel(kwargs.get('xlabel','ANI'))
    if threshold != False:
        plt.xlim([0,2*threshold])

    # Adjust the figure size
    fig = plt.gcf()
    fig.set_size_inches(10,_x_fig_size(len(names),factor=.2))
    plt.subplots_adjust(left=0.3)

    # Adjust the x labels
    plt.tick_params(axis='both', which='major', labelsize=8)
    axes = plt.gca()
    labels = axes.xaxis.get_majorticklocs()
    for i, label in enumerate(labels):
        labels[i] = (1 - float(label)) * 100
    axes.set_xticklabels(labels)

def _make_special_dendrogram(linkage, names, **kwargs):
    '''
    Make the dendrogram used in plot 2

    names can be gotten like:
        db = db.pivot("reference","querry","ani")
        names = list(db.columns)

    Args:
        linkage: result of scipy.cluster.hierarchy.linkage
        names: names of the linkage

    Kwargs:
        name2cluster: dict
        self_thresh: x-axis for soft line
        threshold: x-axis for hard line
        title_sting: title of the plot
        subtitle_string: subtitle of the plot
        winners: list of "winning" genomes (to be marked with star)
        genome2taxonomy: dictionary to add taxonomy information

    Returns:
        Matplotlib primed with a figure
    '''
    # Load possible kwargs
    name2cluster = kwargs.get('name2cluster',False)
    self_thresh = kwargs.get('self_thresh',False)
    threshold = kwargs.get('threshold',False)
    title_string = kwargs.get('title_string','')
    subtitle_string = kwargs.get('subtitle_string','')
    winners = kwargs.get('winners',False)
    genome2taxonomy = kwargs.get('genome2taxonomy',False)

    # Make special things
    if name2cluster != False:
        name2color = gen_color_dictionary(names, name2cluster)
    else:
        name2color = False

    # Make the dendrogram
    sns.set_style('whitegrid')
    g = fancy_dendrogram(linkage,names,name2color,threshold=threshold,self_thresh =\
                        self_thresh)

    # Add the title and subtitle
    plt.suptitle(title_string, y=1, fontsize=18)
    plt.title(subtitle_string, fontsize=10)

    # Adjust the x-axis
    plt.xlabel('Average Nucleotide Identity (ANI)')
    if threshold != False:
        plt.xlim([0,3*threshold])
    plt.tick_params(axis='both', which='major', labelsize=12)
    axes = plt.gca()
    labels = axes.xaxis.get_majorticklocs()
    for i, label in enumerate(labels):
        labels[i] = (1 - float(label)) * 100
    axes.set_xticklabels(labels)
    plt.gca().yaxis.grid(False)

    # Adjust the figure size
    fig = plt.gcf()
    fig.set_size_inches(10,_x_fig_size(len(names),factor=.5))
    plt.subplots_adjust(left=0.5)

    # Mark winning ones
    if type(winners) is not bool:
        ax = plt.gca()
        labels = [item.get_text() for item in ax.get_yticklabels()]
        for i, label in enumerate(labels):
            if label in winners: labels[i] = label + ' *'
        ax.set_yticklabels(labels)

    # Add taxonomy
    if genome2taxonomy != False:
        g2t = genome2taxonomy
        axes = plt.gca()
        labels = [item.get_text() for item in axes.get_yticklabels()]
        for i, label in enumerate(labels):
            labels[i] = "{0}\n{1}".format(label, g2t[label.replace(' *','')])
        axes.set_yticklabels(labels)

def _get_toshow(Gdb):
    '''
    From Gdb, figure out what columns you can show.
    '''
    cols = list(Gdb.columns)
    cols.remove('genome')
    return cols

def _make_scoring_plot(db, bars,**kwargs):
    '''
    Used by winner plot

    db is the database to plot- must contain 'genome' and all columns listed in 'bars'
    bars is all of the columns in the database to become bars
    for taxonomy, put genome2taxonomy in kwargs
    '''
    sns.reset_orig()

    # Make the normalized bar plot
    nd = normalize(db)
    d = pd.melt(nd, id_vars=['genome'], value_vars=bars)
    g = sns.barplot(data=d, y='genome', x='value', hue='variable')

    # Get a list of the un-normalized values
    x = pd.melt(db, id_vars=['genome'], value_vars=bars)
    vals = []
    for variable in x['variable'].unique():
        vals += [v for v in x['value'][x['variable'] == variable].tolist()]

    # Add un-normalized values to barplots
    i = 0
    for p in g.patches:
        g.annotate("{0:.1f}".format(vals[i]), (p.get_width(), p.get_y()+(p.get_height()/1.1) ), fontsize=8)
        i += 1

    # Add taxonomy if available
    axes = plt.gca()
    labels = [item.get_text() for item in axes.get_yticklabels()]
    if kwargs.get('genome2taxonomy',False) != False:
        g2t = kwargs.get('genome2taxonomy')
        for i, label in enumerate(labels):
            labels[i] = "{0}\n{1}".format(label, g2t[label.replace(' *','')])
        axes.set_yticklabels(labels)

    # Adjust labels
    plt.xlabel('Normalized Score')
    plt.legend(loc='lower right')
    plt.tick_params(axis='both', which='major', labelsize=8)

    # Adjust figure size
    fig = plt.gcf()
    fig.set_size_inches(12,_x_fig_size(len(labels), factor=1))
    plt.subplots_adjust(left=0.5)

def _make_mds_plot(name, dist, names, **kwargs):
    '''
    Use MDS to cluster points.

    Based on:
    http://baoilleach.blogspot.com/2014/01/convert-distance-matrix-to-2d.html

    Args:
        name: title of plot
        dist: linkage databases
        names: list of names in linkage database

    Kwargs:
        annotate: if True, write names of all points
        colors: list of colors to use
        shepard: if True, make shepard plot
        tick_spacing: default = .01
        c2c: cluster to color

    Returns:
        Primes plot in matplotlib
    '''

    # load kwargs
    annotate = kwargs.get('annotate', False)
    colors = kwargs.get('colors', False)
    shepard = kwargs.get('shepard', False)
    tick_spacing = kwargs.get('tick_spacing', .01)
    c2c = kwargs.get('cluster2color', False)

    # perform MDS
    mds = manifold.MDS(n_components=2, dissimilarity="precomputed", random_state=6)
    results = mds.fit(dist)
    coords = results.embedding_

    # make shepard plot
    if shepard:
        _shepard_plot(coords, dist, names)

    # plot
    sns.set_style('whitegrid')
    plt.subplots_adjust(bottom = 0.1)
    plt.scatter(
        coords[:, 0], coords[:, 1], marker = 'o', linestyle = 'None',\
        color = colors
        )

    if annotate:
        for label, x, y in zip(names, coords[:, 0], coords[:, 1]):
            plt.annotate(
                label,
                xy = (x, y), xytext = (-20, 20),
                textcoords = 'offset points', ha = 'right', va = 'bottom',
                bbox = dict(boxstyle = 'round,pad=0.5', fc = 'yellow', alpha = 0.5),
                arrowprops = dict(arrowstyle = '->', connectionstyle = 'arc3,rad=0'))

    plt.title('Primary cluster {0}; grid spacing: {1}% ANI'.format(name, tick_spacing*100))

    plt.axis('equal')
    ax = plt.gca()
    ax.set_xticklabels([])
    ax.set_yticklabels([])
    ax.xaxis.set_major_locator(ticker.MultipleLocator(tick_spacing))
    ax.yaxis.set_major_locator(ticker.MultipleLocator(tick_spacing))

    # Add legend
    legendC = c2c
    if legendC != None:
        patches = [mpatches.Patch(color = color, label = label)
                    for label, color in zip(legendC.keys(), legendC.values())]
        lgd = plt.legend(patches, legendC.keys(), loc = 'center left', \
                   bbox_to_anchor=(1, 0.5), prop = {'size': 10, 'style': 'italic'})

def _shepard_plot(coords, dist, names):
    '''
    A componant of the MDS plot
    '''
    table = {'ani_dist':[], 'mds_dist':[]}
    for v1, mx, my in zip(names, coords[:, 0], coords[:, 1]):
        for v2, mx2, my2 in zip(names, coords[:, 0], coords[:, 1]):
            m_dist = calc_dist(mx, my, mx2, my2)

            a_dist = dist.ix[v1, v2]

            table['mds_dist'].append(m_dist)
            table['ani_dist'].append(a_dist)

    adb = pd.DataFrame(table)
    sns.regplot(data=adb, x='ani_dist', y='mds_dist')


'''
****************************    DEPREICATED   *******************************

*   This is where that re-cluster stuff used to be

################################################################################
'''

def cluster_test_wrapper(wd, **kwargs):
    '''
    DEPRICATED
    '''
    # Validate arguments
    cluster = kwargs.get('cluster')
    comp_method = kwargs.get('clustering_method','ANIn')
    if comp_method not in ['ANIn','gANI']:
        raise ValueError()
    clust_method = kwargs.get('clusterAlg')
    threshold = kwargs.pop('threshold',None)
    cov_thresh = float(kwargs.get('minimum_coverage'))
    if threshold != None: threshold = 1- float(threshold)

    # Make a bdb listing the genomes to cluster
    Cdb = wd.get_db('Cdb')
    Bdb = wd.get_db('Bdb')
    genomes = Cdb['genome'][Cdb['primary_cluster'] == int(cluster)].tolist()
    bdb = Bdb[Bdb['genome'].isin(genomes)]

    # Get taxonomy if applicable
    if 'taxonomy' in Bdb:
        genome2taxonomy = Bdb.set_index('genome')['taxonomy'].to_dict()
        kwargs['genome2taxonomy'] = genome2taxonomy

    # Make the comparison database
    Xdb = drep.d_cluster.compare_utils.compare_genomes(bdb, comp_method, wd, **kwargs)

    # Remove values without enough coverage
    if comp_method == 'ANIn':
        Xdb.loc[Xdb['alignment_coverage'] <= cov_thresh, 'ani'] = 0

    # Make it symmetrical
    drep.d_cluster.add_avani(Xdb)
    #Xdb['av_ani'] = Xdb.apply(lambda row: drep.d_cluster.average_ani (row,Xdb),axis=1)
    Xdb['dist'] = 1 - Xdb['av_ani']
    db = Xdb.pivot("reference","querry","dist")

    # Cluster it
    if threshold == None:
        threshold = float(0)
    cdb, linkage = drep.d_cluster.cluster_utils.cluster_hierarchical(db, linkage_method = clust_method, \
                                                                     linkage_cutoff = threshold)

    # Make the plot
    names = list(db.columns)
    if comp_method == 'ANIn':
        kwargs['self_thresh'] = get_highest_self(Xdb, names)
    kwargs['threshold'] = threshold
    kwargs['title_string']="Primary_cluster_{0}_{1}".format(cluster,clust_method)
    kwargs['subtitle_string'] = "Comp method: {0}    ".format(comp_method) +\
            "Clust method: {0}    Min cov: {1}".format(clust_method, cov_thresh)
    kwargs['name2cluster'] = cdb.set_index('genome')['cluster'].to_dict()

    plot_clustertest(linkage, names, wd, **kwargs)

def plot_clustertest(linkage, names, wd, **kwargs):
    '''
    DEPREICATED

    names can be gotten like:
    db = db.pivot("reference","querry","ani")
    names = list(db.columns)
    '''

    # Make the plot directory
    plot_dir = wd.location + '/figures/cluster_tests/'
    if not os.path.exists(plot_dir):
        os.makedirs(plot_dir)

    # Make the dendrogram
    _make_special_dendrogram(linkage,names,**kwargs)

    # Save the dendrogram
    fig = plt.gcf()
    plt.savefig("{0}{1}.pdf".format(plot_dir, kwargs['title_string']))
    plt.show()
    plt.close(fig)



#!/usr/bin/env python3
'''
d_filter - a subset of drep

Filter genomes based on genome length or quality. Also can run prodigal and checkM
'''

import logging
import glob
import gzip
import pandas as pd
import os
import sys
import shutil
import multiprocessing
from Bio import SeqIO

import drep
import drep.WorkDirectory
import drep.d_cluster
import drep.d_choose
import drep.d_bonus

def d_filter_wrapper(wd, **kwargs):
    '''
    Controller for the dRep filter operation

    Args:
        wd (WorkDirectory): The current workDirectory
        **kwargs: Command line arguments

    Keyword Args:
        genomes: genomes to filter in .fasta format
        genomeInfo: location of .csv file with the columns: ["genome"(basename of .fasta file of that genome), "completeness"(0-100 value for completeness of the genome), "contamination"(0-100 value of the contamination of the genome)]

        processors: Threads to use with checkM / prodigal
        overwrite: Overwrite existing data in the work folder
        debug: If True, make extra output when running external scripts

        length: minimum genome length when filtering
        completeness: minimum genome completeness when filtering
        contamination: maximum genome contamination when filtering

        ignoreGenomeQuality: Don't run checkM or do any quality-based filtering (not recommended)
        checkM_method: Either lineage_wf (more accurate) or taxonomy_wf (faster)

    Returns:
        Nothing: stores Bdb.csv, Chdb.csv, and GenomeInfo.csv in the work directory
    '''
    # Load the WorkDirectory.
    logging.debug("Loading work directory in filter")
    workDirectory = drep.WorkDirectory.WorkDirectory(wd)
    logging.debug(str(workDirectory))
    wd = workDirectory

    # Validate arguments
    logging.debug("Validating filter arguments")
    _validate_arguments(workDirectory, **kwargs)

    # Get the thing to filter - bdb (columns = genome, location)
    if wd.hasDb('Bdb'):
        logging.info("Will filter Bdb")
        bdb = wd.get_db('Bdb')
    else:
        logging.info("Will filter the genome list")
        bdb = drep.d_cluster.utils.load_genomes(kwargs['genomes'])

    # Quick sanity check
    sanity_check(bdb)

    # Calculate the length and N50 of all genomes
    logging.info("Calculating genome info of genomes")
    Gdb = calc_genome_info(bdb['location'].tolist())

    # Filter by size if need be
    if kwargs.get('length', 0) > 1:
        logging.debug("Filtering genomes by size")
        bdb = _filter_bdb_length(bdb, Gdb, kwargs['length'])

    # Get comp/con information
    if kwargs.get('ignoreGenomeQuality', False):
        logging.debug("Skipping all quality-based filtering")

    else:
        Gdb = _get_run_genomeInfo(wd, bdb, **kwargs)

    # Filter
    if not kwargs.get('ignoreGenomeQuality', False):
        logging.debug("Filtering genomes")
        bdb = filter_bdb(bdb, Gdb, **kwargs)

    # Save Bdb and genomeInfo
    logging.debug("Storing resulting files")

    workDirectory.store_db(bdb, 'Bdb')
    if not kwargs.get('ignoreGenomeQuality', False):
        workDirectory.store_db(Gdb, 'genomeInfo')

def sanity_check(bdb, **kwargs):
    """
    Make sure there are no duplicate names or anything

    Args:
        bdb:

    Returns:

    """
    # Check for duplicated genome names
    dupped = [g for g, c in bdb['genome'].value_counts().to_dict().items() if c > 1]
    if len(dupped) > 0:
        logging.error("You have duplicate genome basenames! This doesn't work for dRep; each basename needs to be unique. Here are you problem genomes:")
        logging.error(bdb[bdb['genome'].isin(dupped)].sort_values('genome'))
        raise Exception

    # Report number of input genomes
    logging.info(f"{len(bdb):,} genomes were input to dRep")

    # Make a suggestion about greedy?
    pc =  kwargs.get('primary_chunksize', None)
    if pc is not None:
        mc = kwargs.get('multiround_primary_clustering', False)
        if (len(bdb) > pc) & (mc == False):
            logging.info(f"Hey! You have {len(bdb)} genomes and arent using greedy algorithms. See https://drep.readthedocs.io/en/latest/choosing_parameters.html#using-greedy-algorithms for info on how to make your run faster")

    # If you mc, try and tell them to turn on tc
    mc = kwargs.get('multiround_primary_clustering', False)
    tc = kwargs.get('run_tertiary_clustering', False)
    if mc and not tc:
        logging.info(f"Hey! You're running multiround_primary_clustering but not run_tertiary_clustering! You should add --run_tertiary_clustering when running with multiround_primary_clustering to avoid weird placement; see https://drep.readthedocs.io/en/latest/choosing_parameters.html#using-greedy-algorithms for more info")


def _get_run_genomeInfo(workDirectory, bdb, **kwargs):
    '''
    Through kwargs and the wd, get genomeInfo

    Args:
        wd: workDirectory
        bdb: current bdb
        kwrags: keyword arguments

    Keyword arguments:
        no_run: if True, don't run any programs and return None if genomeInfo can't be made

    Returns:
        DataFrame: genomeInfo
    '''

    if kwargs.get('genomeInfo', None) != None:
        logging.debug("Loading provided genome quality information")
        try:
            Idb = pd.read_csv(kwargs.get('genomeInfo'))
        except:
            Idb = pd.read_csv(kwargs.get('genomeInfo'), sep='\t')
        try:
            Tdb = _validate_genomeInfo(Idb, bdb)
        except IndexError:
            bdb_cleanup = bdb.loc[~(bdb['genome'].isin(Idb['genome']))]
            missing_count = bdb_cleanup.shape[0]
            logging.debug("Missing info on {0} genomes,"
                          " running CheckM".format(missing_count))
            Chdb_cleanup = _run_checkM_wrapper(bdb_cleanup,
                                               workDirectory,
                                               **kwargs)
            Idb_cleanup = chdb_to_genomeInfo(Chdb_cleanup)
            Idb_combined = pd.concat([Idb, Idb_cleanup])
            Tdb = _validate_genomeInfo(Idb_combined, bdb)
        Gdb = _add_lengthN50(Tdb, bdb)

    elif workDirectory.hasDb('genomeInfo'):
        logging.debug("Loading genomeInfo.csv from work directory")
        Idb = workDirectory.get_db('genomeInfo')
        Tdb = _validate_genomeInfo(Idb, bdb)
        Gdb = _add_lengthN50(Tdb, bdb)

    elif workDirectory.hasDb('Chdb'):
        logging.debug("Loading Chdb.csv from work directory")
        Chdb = workDirectory.get_db('Chdb')
        Idb = chdb_to_genomeInfo(Chdb)
        Tdb = _validate_genomeInfo(Idb, bdb)
        Gdb = _add_lengthN50(Tdb, bdb)

    elif kwargs.get('no_run', False):
        logging.debug("Making basic genomeInfo")
        Gdb = calc_genome_info(bdb['location'].tolist())
        del Gdb['location']

    else:
        logging.debug("Running CheckM")
        Chdb = _run_checkM_wrapper(bdb, workDirectory, **kwargs)
        Idb = chdb_to_genomeInfo(Chdb)
        Tdb = _validate_genomeInfo(Idb, bdb)
        Gdb = _add_lengthN50(Tdb, bdb)

    return Gdb

def _validate_genomeInfo(Idb, bdb):
    '''
    Validate genomeinfo file; merge in Gdb if needed

    Args:
        genomeInfo: DataFrame of genome info
        bdb: DataFrame with genomes to confirm

    Returns:
        DataFrame: Validated genomeInfo
    '''
    try:
        logging.debug("HERE IS GENOME INFO:")
        logging.debug('\n' + str(Idb.head()))
        logging.debug("There are the columns: {0}".format(list(Idb.columns)))
    except:
        logging.warning(Idb)

    # Make sure it has required columns
    for r in ['completeness', 'contamination', 'genome']:
        if r not in Idb.columns:
            raise KeyError("{0} missing from GenomeInfo".format(r))

    # Make sure correct datatypes
    for r in ['completeness', 'contamination', 'strain_heterogeneity']:
        if r in Idb:
            Idb[r] = Idb[r].astype(float)

    # See if full path was used
    for genome in list(bdb['genome'].unique()):
        if genome not in Idb['genome'].tolist():
            if os.path.basename(genome) in [os.path.basename(x) for x in Idb['genome']]:
                logging.warning("Provided genome info has full genome path- correcting")
                Idb['location'] = Idb['genome']
                Idb['genome'] = [os.path.basename(x) for x in Idb['location']]
                break

    # Make sure it matchs up with bdb
    for genome in list(bdb['genome'].unique()):
        if genome not in Idb['genome'].tolist():
            raise IndexError("{0} missing from GenomeInfo".format(genome))

    # Throw warnings if you think this is weird
    for r in ['completeness', 'contamination', 'strain_heterogeneity']:
        if r in Idb:
            if (Idb[r].max() <= 1) & (Idb[r].min() > 0):
                logging.warning("GenomeInfo has no values over 1 for {0}".format(r) + \
                    "- these should be 0-100, not 0-1!")

    return Idb

def _add_lengthN50(Idb, bdb):
    '''
    If Gdb doesn't have length or N50, add it
    '''
    # Figure out if you need to add them
    add = False
    for r in ['length', 'N50']:
        if r not in Idb:
            add = True

    if not add:
        return Idb

    Gdb = calc_genome_info(bdb['location'].tolist())
    for r in ['length', 'N50']:
        if r not in Idb:
            Idb[r] = Idb['genome'].map(Gdb.set_index('genome')[r].to_dict())

    return Idb

def calc_genome_info(genomes: list):
    '''
    Calculate the length and N50 of a list of genome locations

    Args:
        genomes: list of locations of genomes

    Returns:
        DataFrame: pandas dataframe with ["location", "length", "N50", "genome"]
    '''
    table = {}
    for att in ['location', 'length', 'N50', 'genome']:
        table[att] = []

    for loc in genomes:
        table['location'].append(loc)
        table['genome'].append(os.path.basename(loc))
        table['length'].append(calc_fasta_length(loc))
        table['N50'].append(calc_n50(loc))

    return pd.DataFrame(table)


def filter_bdb(bdb, Gdb, **kwargs):
    '''
    Filter bdb based on Gdb

    Args:
        bdb: DataFrame with ["genome"]
        Gdb: DataFrame with ["genome", "completeness", "contamination"]

    Keyword args:
        min_comp: Minimum genome completeness (%)
        max_con: Maximum genome contamination (%)

    Returns:
        DataFrame: bdb filtered based on completeness and contamination
    '''
    # Get kwargs set up
    min_comp = kwargs.get('completeness',False)
    max_con = kwargs.get('contamination',False)

    # log the number of staring genomes
    start_genomes = list(bdb['genome'].unique())
    if len(start_genomes) <= 0:
        raise ValueError('No genomes')

    # Filter Gdb based on the required metrics
    db = Gdb.copy()
    if min_comp != False:
        db = db[(db['completeness'] >= min_comp)]
    if max_con != False:
        db = db[db['contamination'] <= max_con]

    # Make a list of the remaining genomes
    keep_genomes = list(db['genome'].unique())

    # filter bdb according to these genomes
    bdb = bdb[bdb['genome'].isin(keep_genomes)]

    logging.info("{0:.2f}% of genomes passed checkM filtering".format(\
            (len(keep_genomes)/len(start_genomes))*100))
    return bdb

def _filter_bdb_length(bdb, Gdb, min_length):
    '''
    Filter bdb (['genome', 'location']) based on the minimum length in Gdb
    (['location', 'length', 'N50'])

    Args:
        bdb: DataFrame
        Gdb: DataFrame
        min_length: length to filter on

    Returns:
        DataFrame: bdb filtered by length
    '''

    start = len(bdb['location'].unique())
    bdb['length'] =  bdb['location'].map(Gdb.set_index('location')['length'].to_dict())
    x = bdb[bdb['length'] >= min_length]
    end = len(x['location'].unique())

    logging.info("{0:.2f}% of genomes passed length filtering".format((end/start)*100))

    return x

def validate_chdb(Chdb, bdb):
    '''
    Make sure all genomes in bdb are in Chdb

    Args:
        Chdb: dataframe of checkM information
        bdb: dataframe with ['genome']
    '''
    quit = False
    b_genomes = bdb['genome'].tolist()
    for genome in b_genomes:
        if genome not in Chdb['Bin Id'].tolist():
            logging.error("{0} is not in checkM db".format(genome))
            quit = True
    if quit:
        logging.error("New checkM db needs to be made\nSee https://drep.readthedocs.io/en/latest/advanced_use.html#troubleshooting-checkm for help troubleshooting")
        sys.exit()
    return

def _validate_arguments(wd,**kwargs):
    '''
    Validate arguments to the filter wrapper

    1) Make sure you either have a genome list or Bdb to filter. Bdb is the rarer
    case, but can be done. Throw warnings if needed.

    Args:
        wd (WorkDirectory): The current workDirectory
        **kwargs: Command line arguments

    Returns:
        Nothing
    '''

    # Figure out what you're going to filter
    if wd.hasDb('Wdb'):
        logging.warning("NOTE: Wdb already exists! This will not be filtered! Be sure you know what you're doing")
    if wd.hasDb('Bdb'):
        if kwargs.get('genomes',None) != None:
            logging.error("Both Bdb and a genome list are found- either don't include "\
                    + "a genome list or start a new work directory!")
            sys.exit()
        if wd.hasDb('Cdb'):
            logging.warning("NOTE: Clustering already exists! This will not be filtered! Be sure you know what you're doing")
    else:
        if kwargs.get('genomes',None) == None:
            logging.error("I don't have anything to filter! Give me a genome list")
            sys.exit()

    return

def _run_checkM_wrapper(bdb, workDirectory, **kwargs):
    '''
    Controller for running checkM through a workDirectory
    '''
    # Run prodigal
    prod_folder = workDirectory.get_dir('prodigal')
    logging.info("Running prodigal")
    run_prodigal(bdb['location'].tolist(), prod_folder, wd=workDirectory, **kwargs)

    # Make the folder to house the checkM info
    checkM_loc = workDirectory.get_dir('checkM')

    # Run checkM
    logging.info("Running checkM")
    checkM_outfolder = checkM_loc + 'checkM_outdir/'
    Chdb = run_checkM(prod_folder, checkM_outfolder, wd=workDirectory, **kwargs)
    validate_chdb(Chdb, bdb)

    # Fix genome size and N50 of Chdb
    Chdb = _fix_chdb(Chdb,bdb)

    # Save checkM run
    workDirectory.store_db(Chdb,'Chdb')

    return Chdb

def _fix_chdb(Chdb, Bdb):
    '''
    Add corrent genome length and N50 info when running checkM with proteins
    '''
    g2s = {}
    g2n = {}
    for genome in Chdb['Bin Id'].unique():
        loc = Bdb['location'][Bdb['genome'] == genome].tolist()[0]
        g2s[genome] = calc_fasta_length(loc)
        g2n[genome] = calc_n50(loc)
    Chdb['Genome size (bp)'] = Chdb['Bin Id'].map(g2s)
    Chdb['N50 (scaffolds)'] = Chdb['Bin Id'].map(g2n)

    return Chdb

def calc_n50(loc):
    '''
    Calculate the N50 of a .fasta file

    Args:
        fasta_loc: location of .fasta file.

    Returns:
        int: N50 of .fasta file.
    '''
    lengths = []


    if loc[-3:] == '.gz':
        with gzip.open(loc, "rt") as handle:
            for seq_record in SeqIO.parse(handle, "fasta"):
                lengths.append(len(seq_record))
    else:
        for seq_record in SeqIO.parse(loc, "fasta"):
            lengths.append(len(seq_record))

    # for seq_record in SeqIO.parse(loc, "fasta"):
    #     lengths.append(len(seq_record))

    half = sum(lengths)/2
    tally = 0
    for l in sorted(lengths, reverse=True):
        tally += l
        if tally > half:
            return l

def run_prodigal(genome_list, out_dir, **kwargs):
    '''
    Run prodigal on a set of genomes, store the output in the out_dir

    Args:
        genome_list: list of genomes to run prodigal on
        out_dir: output directory to store prodigal output

    Keyword Args:
        processors: number of processors to multithread with
        exe_loc: location of prodigal excutible (will try and find with shutil if not provided)
        debug: log all of the commands
        wd: if you want to log commands, you also need the wd

    '''
    # Get set up
    t = kwargs.get('processors','6')
    loc = kwargs.get('exe_loc', None)
    if loc == None:
        loc = drep.get_exe('prodigal')

    # Make sure it's a list
    assert type(genome_list) == list

    # Make list of commands
    cmds = []
    for genome in genome_list:
        fna = "{0}{1}".format(os.path.join(out_dir,os.path.basename(genome)),'.fna')
        faa = "{0}{1}".format(os.path.join(out_dir,os.path.basename(genome)),'.faa')
        if os.path.exists(fna) and os.path.exists(faa):
            pass
        else:
            cmds.append(['prodigal','-i',genome,'-d',fna,'-a',faa,'-m','-p','single'])

    # Run commands
    if len(cmds) > 0:
        if ('wd' in kwargs) and (kwargs.get('debug', False) == True):
            logdir = kwargs.get('wd').get_dir('cmd_logs')
        else:
            logdir = False
            #logdir = "/home/mattolm/Programs/drep/tests/test_backend/logs/"

        drep.thread_cmds(cmds, shell=False, logdir=logdir, t=int(t))

    else:
        logging.info("Past prodigal runs found- will not re-run")

def run_checkM(genome_folder_whole, checkm_outf_whole, **kwargs):
    '''
    Run checkM

    WARNING- this will result in wrong genome length and genome N50 estimate, due to
    it being run on prodigal output

    Args:
        genome_folder: location of folder to run checkM on - should be full of files ending in .faa (result of prodigal)
        checkm_outf: location of folder to store checkM output

    Keyword args:
        processors: number of threads
        checkm_method: either lineage_wf or taxonomy_wf
        debug: log all of the commands
        wd: if you want to log commands, you also need the wd
        set_recursion: if not 0, set the python recursion
    '''
    # Get set up
    check_exe = _checkm_get_exe()
    t = str(kwargs.get('processors','6'))
    checkm_method = kwargs.get('checkM_method', 'lineage_wf')
    checkm_group_size = kwargs.get('checkm_group_size', 1000)

    # Set recursion
    R = kwargs.get('set_recursion', '0')
    if R != '0':
        logging.warning('Setting Maximum Recursion depth to {0}'.format(R))
        sys.setrecursionlimit(int(R))

    # Establish groups
    dbs = []
    for genome_folder, checkm_outf in _iterate_checkm_groups(genome_folder_whole, checkm_outf_whole, checkm_group_size):

        # Run checkM initial
        if checkm_method == 'taxonomy_wf':
             cmd = [check_exe,checkm_method,'domain','Bacteria',genome_folder,checkm_outf,'-f',\
                checkm_outf + '/results.tsv','--tab_table','-t',str(t),'-g','-x','faa']
        else:
             cmd = [check_exe,checkm_method,genome_folder,checkm_outf,'-f',\
                checkm_outf + '/results.tsv','--tab_table','-t',str(t),'--pplacer_threads',\
                str(t),'-g','-x','faa']

        logging.debug("Running CheckM with command: {0}".format(' '.join(cmd)))

        if ('wd' in kwargs) & (kwargs.get('debug', False) == True):
            logdir = kwargs.get('wd').get_dir('cmd_logs')
        else:
            logdir = False
        drep.run_cmd(cmd, shell=False, logdir=logdir)

        # Run checkM again for the better table
        if checkm_method == 'taxonomy_wf':
            lineage = checkm_outf + 'Bacteria.ms'
        else:
            lineage = checkm_outf + 'lineage.ms'
        desired_file = checkm_outf + 'Chdb.tsv'
        cmd = [check_exe,'qa', lineage, checkm_outf, '-f', desired_file, '-t',\
                str(t), '--tab_table','-o', '2']
        logging.debug("Running CheckM with command: {0}".format(' '.join(cmd)))

        if ('wd' in kwargs) & (kwargs.get('debug', False) == True):
            logdir = kwargs.get('wd').get_dir('cmd_logs')
        else:
            logdir = False
        drep.run_cmd(cmd, shell=False, logdir=logdir)

        # Load table
        try:
            chdb = pd.read_table(desired_file,sep='\t')
            chdb['Bin Id'] = chdb['Bin Id'].astype(str)
        except:
            logging.error("!!! checkM failed !!!\nSee https://drep.readthedocs.io/en/latest/advanced_use.html#troubleshooting-checkm for help troubleshooting")
            sys.exit()

        # Return table
        dbs.append(chdb)

    return pd.concat(dbs).reset_index(drop=True)

def _checkm_get_exe():
    loc, works = drep.d_bonus.find_program('checkm')
    if loc == None:
        logging.error('Cannot locate the program {0}- make sure its in the system path' \
                      .format('checkm'))
        sys.exit()
    if works == False:
        logging.error('Program {0} is not working!! Im going to crash now' \
                      .format('checkm'))
        sys.exit()
    return loc

def _iterate_checkm_groups(genome_folder_whole, checkm_outf_whole, checkm_group_size):
    """
    Break up the checkM input into groups of checkm_group_size
    """
    # Get the number of prodigal files you have
    total_genes = glob.glob(os.path.join(genome_folder_whole, '*faa'))
    chunks = [total_genes[x:x + checkm_group_size] for x in range(0, len(total_genes), checkm_group_size)]

    if len(chunks) == 1:
        yield genome_folder_whole, checkm_outf_whole

    else:
        logging.info(f"Running checkM in {len(chunks)} chunks")

        if not os.path.isdir(checkm_outf_whole):
            os.mkdir(checkm_outf_whole)

        for i, chunk in enumerate(chunks):
            logging.info(f"Running checkM chunk {i}")

            new_checkm_out = os.path.join(checkm_outf_whole, f"genome_chunk_{i}_out/")
            new_checkm_prods = os.path.join(checkm_outf_whole, f"genome_chunk_{i}_genes")

            os.mkdir(new_checkm_out)
            os.mkdir(new_checkm_prods)

            for faa in chunk:
                os.symlink(faa, os.path.join(new_checkm_prods, os.path.basename(faa)))

            yield new_checkm_prods, new_checkm_out

def calc_fasta_length(fasta_loc):
    '''
    Calculate the length of the .fasta file and retun length

    Args:
        fasta_loc: location of .fasta file

    Returns:
        int: total length of all .fasta files
    '''
    total = 0

    if fasta_loc[-3:] == '.gz':
        with gzip.open(fasta_loc, "rt") as handle:
            for seq_record in SeqIO.parse(handle, "fasta"):
                total += len(seq_record)
    else:
        for seq_record in SeqIO.parse(fasta_loc, "fasta"):
            total += len(seq_record)

    return total

def chdb_to_genomeInfo(chdb):
    '''
    Convert the output of checkM (chdb) into genomeInfo

    Args:
        chdb: dataframe of checkM

    Returns:
        DataFrame: genomeInfo
    '''
    Gdb = chdb.copy()
    Gdb.rename(columns={'Bin Id': 'genome', 'Completeness': 'completeness',\
        'Contamination':'contamination', \
        'Strain heterogeneity':'strain_heterogeneity'}, inplace=True)
    Gdb = Gdb[['genome', 'completeness', 'contamination', 'strain_heterogeneity']]
    return Gdb



import logging
import os
import sys

import numpy as np
import pandas as pd
import scipy.cluster
from scipy.spatial import distance as ssd

import drep.d_cluster.utils

def genome_hierarchical_clustering(Ndb, **kwargs):
    '''
    Cluster ANI database

    Args:
        Ndb: result of secondary clustering

    Keyword arguments:
        clusterAlg: how to cluster the database (default = single)
        S_ani: thershold to cluster at (default = .99)
        cov_thresh: minumum coverage to be included in clustering (default = .5)
        cluster: name of the cluster
        comp_method: comparison algorithm used

    Returns:
        list: [Cdb, {cluster:[linkage, linkage_db, arguments]}]
    '''
    logging.debug('Clustering ANIn database')

    S_Lmethod = kwargs.get('clusterAlg', 'single')
    S_Lcutoff = 1 - kwargs.get('S_ani', .99)
    cov_thresh = float(kwargs.get('cov_thresh',0.5))
    cluster = kwargs.get('cluster','')
    comp_method = kwargs.get('comp_method', 'unk')

    Table = {'genome':[],'secondary_cluster':[]}

    # Handle the case where there's only one genome
    if len(Ndb['reference'].unique()) == 1:
        Table['genome'].append(os.path.basename(Ndb['reference'].unique().tolist()[0]))
        Table['secondary_cluster'].append("{0}_0".format(cluster))
        cluster_ret = []

    else:
        # Make linkage Ndb
        Ldb = drep.d_cluster.utils.make_linkage_Ndb(Ndb, **kwargs)

        # 3) Cluster the linkagedb
        Gdb, linkage = cluster_hierarchical(Ldb, linkage_method= S_Lmethod, \
                                    linkage_cutoff= S_Lcutoff)

        # 4) Extract secondary clusters
        for clust, d in Gdb.groupby('cluster'):
            for genome in d['genome'].tolist():
                Table['genome'].append(genome)
                Table['secondary_cluster'].append("{0}_{1}".format(cluster,clust))

        # 5) Save the linkage
        arguments = {'linkage_method':S_Lmethod,'linkage_cutoff':S_Lcutoff,\
                    'comparison_algorithm':comp_method,'minimum_coverage':cov_thresh}
        cluster_ret = [linkage, Ldb, arguments]

    # Return the database
    Gdb = pd.DataFrame(Table)
    Gdb['threshold'] = S_Lcutoff
    Gdb['cluster_method'] = S_Lmethod
    Gdb['comparison_algorithm'] = comp_method

    return Gdb, cluster_ret


def iteratre_clusters(Bdb, Cdb, id='primary_cluster'):
    '''
    An iterator: Given Bdb and Cdb, yeild smaller Bdb's in the same cluster

    Args:
        Bdb: [genome, location]
        Cdb: [genome, id]
        id: what to iterate on (default = 'primary_cluster')

    Returns:
        list: [d(subset of b), cluster(name of cluster)]
    '''
    Bdb = pd.merge(Bdb,Cdb)
    for cluster, d in Bdb.groupby(id):
        yield d, cluster


def cluster_hierarchical(db, linkage_method= 'single', linkage_cutoff= 0.10):
    '''
    Perform hierarchical clustering on a symmetrical distiance matrix

    Args:
        db: result of db.pivot usually
        linkage_method: passed to scipy.cluster.hierarchy.fcluster
        linkage_cutoff: distance to draw the clustering line (default = .1)

    Returns:
        list: [Cdb, linkage]
    '''
    # Save names
    names = list(db.columns)

    # Generate linkage dataframe
    arr =  np.asarray(db)
    try:
        arr = ssd.squareform(arr)
    except:
        logging.error("The database passed in is not symmetrical!")
        logging.error(arr)
        logging.error(names)
        sys.exit()
    linkage = scipy.cluster.hierarchy.linkage(arr, method= linkage_method)

    # Form clusters
    fclust = scipy.cluster.hierarchy.fcluster(linkage,linkage_cutoff, \
                    criterion='distance')
    # Make Cdb
    Cdb = drep.d_cluster.utils._gen_cdb_from_fclust(fclust,names)

    return Cdb, linkage


#!/usr/bin/env python3

import logging
import glob
import pandas as pd
import os
import shutil
import numpy as np
import pickle
import sys

import drep.WorkDirectory
import drep as dm
import drep.d_cluster as dClust
import drep.d_cluster.cluster_utils
import drep.d_cluster.compare_utils
import drep.d_cluster.controller
import drep.d_filter as d_filter
import drep.d_choose as dChoose

def d_adjust_wrapper(wd,**kwargs):
    # Load the WorkDirectory.
    logging.info("Loading work directory")
    wd = drep.WorkDirectory.WorkDirectory(wd)
    logging.debug(str(wd))

    if kwargs.get('cluster') != None:
        logging.info('adjusting cluster {0}'.format(kwargs.get('cluster')))
        adjust_cluster_wrapper(wd, **kwargs)

    elif kwargs.get('rm_cluster') != None:
        logging.info('removing cluster(s) {0}'.format(' '.join(kwargs.get('rm_cluster'))))
        remove_cluster_wrapper(wd, **kwargs)

def remove_cluster_wrapper(wd, **kwargs):
    # Get information
    Cdb = wd.get_db('Cdb')
    Wdb = wd.get_db('Wdb')

    for cluster in kwargs.get('rm_cluster'):
        type = cluster_type(cluster)

        if type == 'primary_cluster':
            # Make sure cluster is in Cdb
            if int(cluster) not in Cdb[type].tolist():
                logging.error("{0} is not in Cdb- quitting".format(cluster))
                sys.exit()

            # Make sure cluster is Wdb
            W_Pclusters = [x.split('_')[0] for x in Wdb['cluster'].tolist()]
            if cluster not in W_Pclusters:
                logging.error("{0} is not in Wdb- quitting".format(cluster))
                sys.exit()

            # Remove the cluster
            remove_primary_cluster(cluster, wd, **kwargs)

        elif type == 'secondary_cluster':

            # Make sure cluster is in Cdb
            if cluster not in Cdb[type].tolist():
                logging.error("{0} is not in Cdb- quitting".format(cluster))
                sys.exit()

            # Make sure cluster is in Wdb
            if cluster not in Wdb['cluster'].tolist():
                logging.error("{0} is not in Wdb- quitting".format(cluster))
                sys.exit()

            # Remove the cluster
            remove_secondary_cluster(cluster, wd, **kwargs)


def remove_primary_cluster(Rcluster, wd, **kwargs):
    # Get things to alter
    Cdb = wd.get_db('Cdb')
    Wdb = wd.get_db('Wdb')

    # If this is a singleton, just remove the associated secondary cluster
    if len(Cdb['genome'][Cdb['primary_cluster'] == int(Rcluster)].unique()) == 1:
        Rsecondary = "{0}_0".format(Rcluster)
        logging.error("{0} is a primary singleton- will remove {1} instead".format(Rcluster, Rsecondary))
        remove_secondary_cluster(Rsecondary, wd, **kwargs)
        return

    # Find the genome(s) to remove
    Wdb['pclust'] = [x.split('_')[0] for x in Wdb['cluster'].tolist()]
    Rgenomes = Wdb['genome'][Wdb['pclust'] == Rcluster].tolist()
    for Rgenome in Rgenomes:
        if not os.path.isfile('{0}/dereplicated_genomes/{1}'.format(wd.location, Rgenome)):
            raise ValueError()

    # Find the pickle to remove
    Rpickle = "{0}/data/Clustering_files/secondary_linkage_cluster_{1}.pickle".format(\
            wd.location, Rcluster)
    if not os.path.isfile(Rpickle):
        raise ValueError()

    logging.info("will remove cluster {0}, genomes {1}, and pickle {2}".format(Rcluster, \
            ' '.join(Rgenomes), os.path.basename(Rpickle)))

    # Remove cluster from Wdb
    newWdb = Wdb[Wdb['pclust'] != Rcluster]
    del newWdb['pclust']

    # Remove cluster from Cdb
    newCdb = Cdb[Cdb['primary_cluster'] != int(Rcluster)]

    # Remove genomes
    for Rgenome in Rgenomes:
        os.remove('{0}/dereplicated_genomes/{1}'.format(wd.location, Rgenome))

    # Remove pickle
    os.remove(Rpickle)

    # Save removed stuff
    wd.store_db(newCdb,'Cdb',overwrite=True)
    wd.store_db(newWdb,'Wdb',overwrite=True)

    logging.info("done")


def remove_secondary_cluster(Rcluster, wd, **kwargs):
    # Get things to alter
    Cdb = wd.get_db('Cdb')
    Wdb = wd.get_db('Wdb')

    # Find the genome to remove
    Rgenome = Wdb['genome'][Wdb['cluster'] == Rcluster].tolist()[0]
    if not os.path.isfile('{0}/dereplicated_genomes/{1}'.format(wd.location, Rgenome)):
        raise ValueError('Bug')

    logging.info("will remove {0} and genome {1}".format(Rcluster, Rgenome))

    # Remove cluster from Wdb
    newWdb = Wdb[Wdb['cluster'] != Rcluster]

    # Remove cluster from Cdb
    newCdb = Cdb[Cdb['secondary_cluster'] != Rcluster]

    # Remove genome
    os.remove('{0}/dereplicated_genomes/{1}'.format(wd.location, Rgenome))

    # Save removed stuff
    wd.store_db(newCdb,'Cdb',overwrite=True)
    wd.store_db(newWdb,'Wdb',overwrite=True)

    logging.info("done")

def cluster_type(cluster):
    if '_' in cluster:
        return 'secondary_cluster'
    else:
        return 'primary_cluster'

def adjust_cluster_wrapper(wd, **kwargs):
    # Validate arguments
    cluster = kwargs.get('cluster')
    if cluster == None:
        logging.error("Must specify a cluster")
        sys.exit()
    comp_method = kwargs.get('clustering_method')
    clust_method = kwargs.get('clusterAlg')
    threshold = kwargs.pop('threshold',None)
    cov_thresh = float(kwargs.get('minimum_coverage'))
    if threshold != None: threshold = 1- float(threshold)

    # Make a bdb listing the genomes to cluster
    Cdb = wd.get_db('Cdb')
    Bdb = wd.get_db('Bdb')
    genomes = Cdb['genome'][Cdb['primary_cluster'] == int(cluster)].tolist()
    bdb = Bdb[Bdb['genome'].isin(genomes)]

    # Make the comparison database
    Xdb = drep.d_cluster.compare_utils.compare_genomes(bdb, comp_method, wd, **kwargs)

    # Remove values without enough coverage
    Xdb.loc[Xdb['alignment_coverage'] <= cov_thresh, 'ani'] = 0

    # Make it symmetrical
    Xdb['av_ani'] = Xdb.apply(lambda row: dClust.average_ani (row,Xdb),axis=1)
    Xdb['dist'] = 1 - Xdb['av_ani']
    db = Xdb.pivot("reference","querry","dist")

    # Cluster it
    if threshold == None:
        threshold = float(0)
    cdb, linkage = drep.d_cluster.cluster_utils.cluster_hierarchical(db, linkage_method = clust_method, \
                                                                     linkage_cutoff = threshold)

    # Save the pickle
    data_folder = wd.location + '/data/Clustering_files/'
    arguments = {'linkage_method':clust_method,'linkage_cutoff':threshold,\
                        'comparison_algorithm':comp_method,'minimum_coverage':cov_thresh}
    pickle_name = "secondary_linkage_cluster_{0}.pickle".format(cluster)
    logging.debug('Saving secondary_linkage pickle {1} to {0}'.format(data_folder,\
                                                        pickle_name))
    with open(data_folder + pickle_name, 'wb') as handle:
        pickle.dump(linkage, handle)
        pickle.dump(db,handle)
        pickle.dump(arguments,handle)

    # Change the cdb
    cdb['cluster_method'] = clust_method
    cdb['comparison_algorithm'] = comp_method
    cdb['threshold'] = str(threshold)
    cdb['primary_cluster'] = int(cluster)
    cdb['secondary_cluster'] = ["{0}_{1}".format(cluster, x) for x in cdb['cluster']]
    del cdb['cluster']

    # Incorporate it into the existing Cdb
    Cdb = Cdb[Cdb['primary_cluster'] != int(cluster)]
    Cdb = pd.concat([Cdb,cdb], ignore_index=True)

    # Save it
    wd.store_db(Cdb,'Cdb',overwrite=True)

    if (not kwargs.get('skip_winner',False)) and wd.hasDb('Wdb'):

        # Remake Wdb based on Sdb
        oriWdb = wd.get_db('Wdb')
        newWdb = dChoose.pick_winners(wd.get_db('Sdb'),Cdb)
        change = accounce_changes(newWdb, oriWdb)
        wd.store_db(newWdb,'Wdb',overwrite=True)

        if change:
            # Change the ./dereplicated genomes thing
            logging.info("Remaking ./dereplicated genomes...")
            output_folder = wd.location + '/dereplicated_genomes/'
            dm.clobber_dir(output_folder,dry=kwargs.get('dry',False),\
                            overwrite=True)

            for genome in newWdb['genome'].unique():
                loc = Bdb['location'][Bdb['genome'] == genome].tolist()[0]
                shutil.copy2(loc, "{0}{1}".format(output_folder,genome))

def accounce_changes(newWdb, oriWdb):
    changes = False

    for cluster in newWdb['cluster'].unique():
        w1 = newWdb['genome'][newWdb['cluster'] == cluster].unique()[0]
        if cluster not in oriWdb['cluster'].tolist():
            change = "{0} is the winner of the new cluster {1}".format(w1,cluster)
            logging.info(change)
            changes = True
            continue
        w2 = oriWdb['genome'][oriWdb['cluster'] == cluster].unique()[0]
        if w1 != w2:
            change = "the winner of cluster {0} is now {1} (was {2})".format(\
                        cluster, w1, w2)
            logging.info(change)
            changes = True
            continue

    for cluster in oriWdb['cluster'].unique():
        if cluster not in newWdb['cluster'].tolist():
            change = "cluster {0} no longer exists".format(cluster)
            logging.info(change)
            changes = True

    if not changes:
        change = "No secondary clusters were made or destroyed, and none have new winners"
        logging.info(change)

    return changes


def test_adjust():
    print("Write this you lazy bum")

if __name__ == '__main__':
	test_adjust()
